from __future__ import annotations
import typing as t
from .protocol.payload.base_struct import (
    RequestBodyStruct,
    ResponseBodyStruct
)
from .protocol.struct_header import ResponseHeader
from ..token import Token


# The defenition of generic Types

HeaderType = t.TypeVar("HeaderType", bound="ResponseHeader", covariant=True)
T_ = t.TypeVar("T_", covariant=True)


class IsDataclass(t.Protocol):
    __dataclass_fields__: t.Dict[str, t.Any]
    def __init__(self, *args, **kwargs) -> None: ...  # noqa: E704


class ICommand(t.Protocol):
    code: t.ClassVar[int]
    pushed: t.ClassVar[bool]


SetStructType = t.TypeVar("SetStructType", bound="RequestBodyStruct", covariant=True)
GetStructType = t.TypeVar("GetStructType", bound="ResponseBodyStruct", covariant=True)


class ICmdOnlySet(ICommand, t.Protocol):
    """A template class which provide only <cmd_set> method."""
    SetDataAttr: t.Type[SetStructType]  # type: ignore
    set: t.Callable[..., Token[None]]


class ICmdOnlyGet(ICommand, t.Protocol):
    """A template class which provide only <cmd_get> method."""
    GetDataAttr: t.Type[GetStructType]  # type: ignore
    get: t.Callable[[], Token[GetStructType]]  # type: ignore


XoaCommandType = t.Union[ICmdOnlySet, ICmdOnlyGet]



from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.commands import (
    C_RESTPORT,
    C_RESTENABLE,
    C_RESTSTATUS,
    C_RESTCONTROL,
)


class RestApiServer:
    """
    Rest API server for the Valkyrie tester
    """

    def __init__(self, conn: "itf.IConnection") -> None:
        self.port = C_RESTPORT(conn)
        """TCP port for the REST server.
                
        :type: C_RESTPORT
        """

        self.enable = C_RESTENABLE(conn)
        """Enable the REST server.
                
        :type: C_RESTENABLE
        """

        self.status = C_RESTSTATUS(conn)
        """Status of the REST server.
                
        :type: C_RESTSTATUS
        """

        self.control = C_RESTCONTROL(conn)
        """Control the REST server.
                
        :type: C_RESTCONTROL
        """



from __future__ import annotations
from functools import (
    cached_property,
    partial,
)

from io import BytesIO
from typing import (
    Any,
    ClassVar,
    Generator,
    Iterator,
    NamedTuple,
    Type,
    cast,
)
from typing_extensions import (
    Self,
    dataclass_transform,
)

from . import utils
from .field import (
    FieldSpecs,
    field,
)
from .descriptor import (
    ResponseFieldDescr,
    RequestFieldDescr
)
from .exceptions import (
    FieldDeclarationError,
    FirmwareVersionError,
)


RESPONSE_CLS_NAME = "ResponseBodyStruct"
SKIP_CLASSES = (
    "RequestBodyStruct",
    RESPONSE_CLS_NAME,
)


class CellInfo(NamedTuple):
    fmt: str
    offset: int
    bsize: int | None


class Cell:
    __slots__ = (
        "name",
        "spec",
        "_prev",
        "_next"
    )

    def __init__(self, name: str, spec: FieldSpecs) -> None:
        self.name = name
        self.spec = spec

        self._prev: Cell | None = None
        self._next: Cell | None = None

    def info(self, buff: memoryview | None = None, offset: int = 0) -> CellInfo:
        bsize = self.spec.calc_bsize(buff, offset)
        fmt = self.spec.format(bsize)
        return CellInfo(fmt, offset, bsize)

    @property
    def is_dynamic(self) -> bool:
        return self.spec.is_dynamic


class Order:
    __slots__ = (
        "__start",
        "__head",
        "__head_idx",
        "__stencil",
        "__dict__",
    )

    def __init__(self) -> None:
        self.__start: Cell | None = None
        self.__head: Cell | None = None
        self.__head_idx: int = -1
        self.__stencil: tuple[tuple[str, int], ...] = ()

    def __iter__(self):
        temp = self.__start
        while temp is not None:
            yield temp
            temp = temp._next

    def __reversed__(self):
        temp = self.__head
        while temp is not None:
            yield temp
            temp = temp._prev

    @cached_property
    def is_dynamic(self) -> bool:
        """Computed only once, Determinate if structure is contain types with dynamic lenght or not"""
        return any(c.is_dynamic for c in self)

    @property
    def field_names(self) -> Iterator[str]:
        """Get iterator of ordered fields names"""
        return iter(itm.name for itm in self)

    def append(self, name: str, data: FieldSpecs) -> int:
        """Append field structure"""
        self.__head_idx += 1
        c = Cell(name, data)
        if not self.__start:
            self.__start = c
        if self.__head:
            self.__head._next = c
            c._prev = self.__head
        self.__head = c
        return self.__head_idx

    def bake(self) -> None:
        """Bake tuple of the offsets for static length structures, at the defenition time."""
        if self.is_dynamic:
            return None
        self.__stencil = tuple(self.__construct_stencil(None))

    def get_stencil(self, buffer: memoryview) -> tuple[tuple[str, int], ...]:
        """Get offsets for the instance. If struct is static get baked offsets, otherwise compute them for the individual instance"""
        if self.is_dynamic:
            return tuple(self.__construct_stencil(buffer))
        return self.__stencil

    def __construct_stencil(self, buffer: memoryview | None) -> Generator[tuple[str, int], None, None]:
        """An private method for construct stencils items."""
        current_node = self.__start
        cumulative_sum = 0

        while current_node is not None:
            info_tuple = current_node.info(buffer, cumulative_sum)
            yield (info_tuple.fmt, cumulative_sum)
            cumulative_sum += info_tuple.bsize or 0
            current_node = current_node._next


class OrderedMeta(type):
    def __new__(cls: Type[Self], clsname: str, bases: tuple[Type], clsdict: dict[str, Any]) -> Self:
        if clsname not in SKIP_CLASSES:
            is_response = any(iter(cls_.__name__ == RESPONSE_CLS_NAME for cls_ in bases))
            annotations = utils.resolve_annotations(
                clsdict.get('__annotations__', {}),
                clsdict.get('__module__', None)
            )
            clsdict = cls._prepare_order(
                annotations=annotations,
                clsdict=clsdict,
                is_response=is_response
            )
        return super().__new__(cls, clsname, bases, clsdict)

    @classmethod
    def __prepare__(cls, clsname: str, bases: tuple) -> dict[str, Any]:
        return dict()

    @staticmethod
    def _prepare_order(annotations: dict[str, Any], clsdict: dict[str, Any], is_response: bool) -> dict[str, Any]:
        clsdict["_order"] = order = Order()
        for field_name, client_type in annotations.items():
            field_specs = clsdict.get(field_name, None)
            if not isinstance(field_specs, FieldSpecs):
                raise FieldDeclarationError(field_name)
            position_idx = order.append(field_name, field_specs)
            descriptor = ResponseFieldDescr if is_response else RequestFieldDescr
            clsdict[field_name] = descriptor(
                idx=position_idx,
                specs=field_specs,
                user_type=client_type,
            )
        order.bake()
        return clsdict


@dataclass_transform(kw_only_default=True, field_descriptors=(field, FieldSpecs,))  # type: ignore[Pylance, false positive]
class RequestBodyStruct(metaclass=OrderedMeta):
    """Request Body class"""

    __slots__ = ("_buffer", "_order", "__stored", "__nbytes")

    def __init__(self, **kwargs) -> None:
        self._buffer = BytesIO()
        for name in cast(Order, self._order).field_names:
            if name not in kwargs:
                raise AttributeError(f"[{name}] is required!")
            setattr(self, name, kwargs[name])
        self.__stored = kwargs
        self.__nbytes = self._buffer.getbuffer().nbytes
        padding = bytes(4 - (self.__nbytes % 4) if self.__nbytes % 4 else 0)
        self._buffer.write(padding)

    def __repr__(self) -> str:
        cls_name = f"{self.__class__.__qualname__}"
        vals = ", ".join(
            f"{n}={v!r}" for n, v in self.__stored.items()
        )
        return f"{cls_name}({vals})"

    def nbytes(self) -> int:
        """Return number of byffer bytes"""
        return self.__nbytes

    def nbytes_with_padding(self) -> int:
        return self._buffer.getbuffer().nbytes

    def to_dict(self) -> dict[str, Any]:
        """Get Dict representation of the object"""
        return self.__stored

    def to_hex(self) -> str:
        """Get buffer as hex string"""
        return self.to_bytes().hex()

    def to_bytes(self) -> bytes:
        """Get buffer as bytes"""
        return self._buffer.getvalue()


def get_val(inst: object, fn: str) -> Any:
    """
    An helper method for get a value fro mobject and in case if FirmwareVersionError swap the value as not supported.

    IMPORTANT: only used in conversion to tuple, dict
    """
    # TODO: Allow to skip not supported values for tuple and dict
    try:
        val = getattr(inst, fn)
    except FirmwareVersionError:
        return "NOT_SUPORTED_BY_FIRMWARE"
    else:
        return val


class ResponseBodyStruct(metaclass=OrderedMeta):
    """Response Body class"""

    __slots__ = ("_buffer", "_order", "_stencil")
    _order: ClassVar[Order]

    def __init__(self, packet_body: bytes | bytearray | memoryview) -> None:
        self._buffer = memoryview(packet_body).toreadonly()
        self._stencil = self._order.get_stencil(self._buffer)

    def __repr__(self) -> str:
        cls_name = self.__class__.__qualname__
        vals = ", ".join(
            f"{n}={v!r}" for n, v in self.to_dict().items()
        )
        return f"{cls_name}({vals})"

    def nbytes(self) -> int:
        """Return number of byffer bytes"""
        return self._buffer.nbytes

    def to_tuple(self) -> tuple:
        """Get py values as tuple"""
        get_val_ = partial(get_val, self)
        return tuple(map(get_val_, self._order.field_names))

    def to_dict(self) -> dict[str, Any]:
        """Get Dict representation of the object"""
        return {
            name: get_val(self, name)
            for name in self._order.field_names
        }

    def to_hex(self) -> str:
        """Get buffer as hex string"""
        return self._buffer.hex()

    def to_bytes(self) -> bytes:
        """Get buffer as bytes"""
        return self._buffer.tobytes()



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.commands import (
    PR_FLOWTOTAL,
    PR_FLOWCLEAR,
)


class ReceptionStatistics:
    """Chimera RX statistics."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, flow_index: int) -> None:
        self.total = PR_FLOWTOTAL(conn, module_id, port_id, flow_index)
        """RX statistics of a flow.

        :type: PR_FLOWTOTAL
        """

        self.clear = PR_FLOWCLEAR(conn, module_id, port_id, flow_index)
        """Clear RX statistics of a flow.

        :type: PR_FLOWCLEAR
        """



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.commands import (
    P4G_TLS_ENABLE,
    P4G_TLS_CIPHER_SUITES,
    P4G_TLS_MAX_RECORD_SIZE,
    P4G_TLS_CERTIFICATE_FILENAME,
    P4G_TLS_PRIVATE_KEY_FILENAME,
    P4G_TLS_DHPARAMS_FILENAME,
    P4G_TLS_CLOSE_NOTIFY,
    P4G_TLS_ALERT_WARNING_COUNTERS,
    P4G_TLS_ALERT_FATAL_COUNTERS,
    P4G_TLS_STATE_CURRENT,
    P4G_TLS_STATE_TOTAL,
    P4G_TLS_STATE_RATE,
    P4G_TLS_RX_PAYLOAD_COUNTERS,
    P4G_TLS_TX_PAYLOAD_COUNTERS,
    P4G_TLS_RX_PAYLOAD_BYTES_HIST,
    P4G_TLS_TX_PAYLOAD_BYTES_HIST,
    P4G_TLS_HANDSHAKE_HIST,
    P4G_TLS_SERVER_NAME,
    P4G_TLS_PROTOCOL_VER,
    P4G_TLS_MIN_REQ_PROTOCOL_VER,
)


class GCountersTlsState:
    """TLS state counters"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.current = P4G_TLS_STATE_CURRENT(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_STATE_CURRENT
        """
        self.total = P4G_TLS_STATE_TOTAL(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_STATE_TOTAL
        """
        self.rate = P4G_TLS_STATE_RATE(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_STATE_RATE
        """


class GCountersTlsAlert:
    """TLS alert counters"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.warning = P4G_TLS_ALERT_WARNING_COUNTERS(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_ALERT_WARNING_COUNTERS
        """
        self.fatal = P4G_TLS_ALERT_FATAL_COUNTERS(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_ALERT_FATAL_COUNTERS
        """


class GCountersTlsPayload:
    """TLS payload counters"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.rx = P4G_TLS_RX_PAYLOAD_COUNTERS(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_RX_PAYLOAD_COUNTERS
        """
        self.tx = P4G_TLS_TX_PAYLOAD_COUNTERS(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_TX_PAYLOAD_COUNTERS
        """


class GCountersTls:
    """TLS counters"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.alert = GCountersTlsAlert(conn, module_id, port_id, group_idx)
        """TLS alert counters"""
        self.payload = GCountersTlsPayload(conn, module_id, port_id, group_idx)
        """TLS payload counters"""
        self.state = GCountersTlsState(conn, module_id, port_id, group_idx)
        """TLS state counters"""


class GHistogramTlsPayload:
    """TLS payload histogram"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.rx_bytes = P4G_TLS_RX_PAYLOAD_BYTES_HIST(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_RX_PAYLOAD_BYTES_HIST
        """
        self.tx_bytes = P4G_TLS_TX_PAYLOAD_BYTES_HIST(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_TX_PAYLOAD_BYTES_HIST
        """


class GHistogramTls:
    """TLS histogram"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.handshake = P4G_TLS_HANDSHAKE_HIST(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_HANDSHAKE_HIST
        """
        self.payload = GHistogramTlsPayload(conn, module_id, port_id, group_idx)
        """TLS payload histogram"""


class GProtocolTls:
    """TLS protocol version"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.version = P4G_TLS_PROTOCOL_VER(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_PROTOCOL_VER
        """
        self.min_required_version = P4G_TLS_MIN_REQ_PROTOCOL_VER(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_MIN_REQ_PROTOCOL_VER
        """


class GFileTls:
    """TLS certificate and key files"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.certificate_path = P4G_TLS_CERTIFICATE_FILENAME(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_CERTIFICATE_FILENAME
        """
        self.private_key_path = P4G_TLS_PRIVATE_KEY_FILENAME(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_PRIVATE_KEY_FILENAME
        """
        self.dhparams_path = P4G_TLS_DHPARAMS_FILENAME(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_DHPARAMS_FILENAME
        """


class GTls:
    """TLS configuration"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.enable = P4G_TLS_ENABLE(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_ENABLE
        """
        self.server_name = P4G_TLS_SERVER_NAME(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_SERVER_NAME
        """
        self.close_notify = P4G_TLS_CLOSE_NOTIFY(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_CLOSE_NOTIFY
        """
        self.cipher_suites = P4G_TLS_CIPHER_SUITES(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_CIPHER_SUITES
        """
        self.max_record_size = P4G_TLS_MAX_RECORD_SIZE(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TLS_MAX_RECORD_SIZE
        """
        self.file = GFileTls(conn, module_id, port_id, group_idx)
        """Cert and key file"""
        self.protocol = GProtocolTls(conn, module_id, port_id, group_idx)
        """Protocol version"""
        self.counters = GCountersTls(conn, module_id, port_id, group_idx)
        """Counters"""
        self.histogram = GHistogramTls(conn, module_id, port_id, group_idx)
        """Histogram"""



from __future__ import annotations


class XmpException(Exception):
    ...


class FirmwareVersionError(XmpException):
    def __init__(self, cmd_name: str, field_name: str, min_version: int | None) -> None:
        self.cmd_name = cmd_name
        self.field_name = field_name
        self.min_version = min_version
        self.msg = f"The field: [{self.field_name}] of command {self.cmd_name} is available only from the firmware version: {self.min_version}"
        super().__init__(self.msg)


class FieldDeclarationError(TypeError):
    def __init__(self, f_name: str) -> None:
        self.msg = f"Structure Field {f_name!r} must be described with <field method>"
        super().__init__(self.msg)



from __future__ import annotations
from dataclasses import dataclass
from enum import IntEnum
import os
import sys
from typing import List
from ipaddress import (
    IPv4Address,
    IPv6Address,
)
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from xoa_driver.internals.core.transporter.protocol.payload import (  # noqa: E402
    RequestBodyStruct,
    field,
    XmpByte,
    XmpHex,
    XmpInt,
    XmpIPv4Address,
    XmpIPv6Address,
    XmpLong,
    XmpMacAddress,
    XmpSequence,
    XmpShort,
    XmpStr,
    Hex
)


# region TypeTesting

def test_byte() -> None:
    data = b"\xff\xff\x00\x00"
    assert len(data) == 4, "for current test case the data must be of 2 bytes long"

    class SetDataAttr(RequestBodyStruct):
        custom_field: int = field(XmpByte())
        custom_field2: int = field(XmpByte(signed=True))

    obj = SetDataAttr(
        custom_field=255,
        custom_field2=-1
    )
    assert obj.to_bytes() == data


def test_short() -> None:
    data = b"\x02\x05\x00\x00"

    class SetDataAttr(RequestBodyStruct):
        custom_field: int = field(XmpShort())

    obj = SetDataAttr(custom_field=517)
    assert obj.to_bytes() == data


def test_int() -> None:
    data = b"\x00\x00\x00\xff"

    class SetDataAttr(RequestBodyStruct):
        custom_field: int = field(XmpInt(climb=(10, 20)))

    obj = SetDataAttr(custom_field=255)
    assert obj.to_bytes() == data


def test_long() -> None:
    data = b"\x00\x00\x00\xff\x00\x00\x00\xff"

    class SetDataAttr(RequestBodyStruct):
        custom_field: int = field(XmpLong())

    obj = SetDataAttr(custom_field=1095216660735)
    assert obj.to_bytes() == data


def test_hex() -> None:
    data = b"\x00\x00\x00\xff"

    class SetDataAttr(RequestBodyStruct):
        custom_field: Hex = field(XmpHex(size=1))
        custom_field2: Hex = field(XmpHex(size=3))

    obj = SetDataAttr(
        custom_field=Hex("00"),
        custom_field2=Hex("ff")
    )
    assert obj.to_bytes() == data


def test_hex_string() -> None:
    data = b'\x04\xf4\xbc\x9cs\xd1\x04\xf4\xbc\x9cs\xd0\xff\xff\x00\x00'

    class SetDataAttr(RequestBodyStruct):
        hex_data: Hex = field(XmpHex())

    obj = SetDataAttr(hex_data=Hex("04F4BC9C73D104F4BC9C73D0FFFF"))
    assert data == obj.to_bytes()


def test_ip_v4() -> None:
    data = b'\x7f\x00\x00\x01'

    class SetDataAttr(RequestBodyStruct):
        custom_field: IPv4Address = field(XmpIPv4Address())

    obj = SetDataAttr(custom_field=IPv4Address("127.0.0.1"))
    assert obj.to_bytes() == data


def test_ip_v6() -> None:
    data = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab'

    class SetDataAttr(RequestBodyStruct):
        custom_field: IPv6Address = field(XmpIPv6Address())

    obj = SetDataAttr(custom_field=IPv6Address("::ab"))
    assert obj.to_bytes() == data


def test_mac() -> None:
    data = b'\xf4\xd4\x88\x67\xc5\xda\x00\x00'

    class SetDataAttr(RequestBodyStruct):
        custom_field: Hex = field(XmpMacAddress())

    obj = SetDataAttr(custom_field=Hex("f4d48867c5da"))
    assert obj.to_bytes() == data


def test_string() -> None:
    data1 = b'Odin-10G-1S-6P[b]\x00\x00\x00'
    data2 = b"Xena"
    data3 = b""

    class SetDataAttr(RequestBodyStruct):
        custom_field: str = field(XmpStr())

    obj1 = SetDataAttr(custom_field="Odin-10G-1S-6P[b]")
    obj2 = SetDataAttr(custom_field="Xena")
    obj3 = SetDataAttr(custom_field="")

    assert obj1.to_bytes() == data1
    assert obj2.to_bytes() == data2
    assert obj3.to_bytes() == data3


def test_unlimited_list() -> None:
    data = b'\x00\x02\x00\x06\x00\x06\x00\x06'

    class SetDataAttr(RequestBodyStruct):
        port_counts: List[int] = field(XmpSequence(types_chunk=[XmpByte()]))

    obj = SetDataAttr(
        port_counts=[0, 2, 0, 6, 0, 6, 0, 6]
    )
    assert obj.to_bytes() == data


def test_limited_list() -> None:
    data = b'\x00\x02\x00\x00'
    MAX_LEN = 2

    class SetDataAttr(RequestBodyStruct):
        custom_field: List[int] = field(XmpSequence(types_chunk=[XmpByte()], length=MAX_LEN))

    try:
        SetDataAttr(
            custom_field=[0, 2, 0, 6, 0, 6, 0, 6]
        )
    except Exception:
        pass
    else:
        assert False, "Some how passed longer list then expected"

    obj = SetDataAttr(
        custom_field=[0, 2]
    )

    assert obj.to_bytes() == data


# endregion


# region TypeConversion

class UsersEnum(IntEnum):
    OPTION_1 = 1
    OPTION_2 = 2
    OPTION_3 = 3
    OPTION_4 = 5


def test_enum_global_def() -> None:
    data = b"\x02\x05\x00\x00"

    class SetDataAttr(RequestBodyStruct):
        custom_field: UsersEnum = field(XmpByte())
        custom_field2: UsersEnum = field(XmpByte())

    obj = SetDataAttr(
        custom_field=UsersEnum.OPTION_2,
        custom_field2=UsersEnum.OPTION_4
    )
    assert obj.to_bytes() == data


def test_enum_local_def() -> None:
    """
    At the Moment the defenition of the enum under the local namespace will faile
    """

    class UsersLocalEnum(IntEnum):
        OPTION_1 = 1
        OPTION_2 = 2
        OPTION_3 = 3
        OPTION_4 = 5

    class SetDataAttr(RequestBodyStruct):
        custom_field: UsersLocalEnum = field(XmpByte())
        custom_field2: UsersLocalEnum = field(XmpByte())

    try:
        SetDataAttr(
            custom_field=UsersLocalEnum.OPTION_2,
            custom_field2=UsersLocalEnum.OPTION_4
        )
    except TypeError as e:
        assert str(e) == "'ForwardRef' object is not callable"
    else:
        assert False, "Magicly it start to works"


def test_bool() -> None:
    data = b"\x01\x00\x00\x00"

    class SetDataAttr(RequestBodyStruct):
        custom_field: bool = field(XmpByte())
        custom_field2: bool = field(XmpByte())

    obj = SetDataAttr(
        custom_field=True,
        custom_field2=False
    )
    assert obj.to_bytes() == data


class OnOff(IntEnum):
    OFF = 0
    ON = 1


@dataclass
class ArpChunk:
    ipv4_address: IPv4Address
    prefix: int
    patched_mac: OnOff
    mac_address: Hex


def test_chunk_type() -> None:
    data = b'\x7f\x00\x00\x01\x02\x05\x01\xf4\xd4\x88\x67\xc5\xda\x7f\x00\x00\x03\x02\x05\x00\xf4\xd4\x88\x55\xc5\xda\x00\x00'

    class SetDataAttr(RequestBodyStruct):
        custom_field: List[ArpChunk] = field(XmpSequence(types_chunk=[XmpIPv4Address(), XmpShort(), XmpByte(), XmpMacAddress()]))

    obj = SetDataAttr(
        custom_field=[
            ArpChunk(
                ipv4_address=IPv4Address("127.0.0.1"),
                prefix=517,
                patched_mac=OnOff.ON,
                mac_address=Hex("f4d48867c5da")
            ),
            ArpChunk(
                ipv4_address=IPv4Address("127.0.0.3"),
                prefix=517,
                patched_mac=OnOff.OFF,
                mac_address=Hex("f4d48855c5da")
            )
        ]
    )
    assert obj.to_bytes() == data

# endregion



from __future__ import annotations
import struct

from . import _constants as const
from . import _utils
from .struct_header import ProtocolHeader
from .payload import RequestBodyStruct

NOT_SET_IDENTIFIER = 0


class Request:
    __slots__ = (
        "class_name",
        "header",
        "index_values",
        "values"
    )

    def __init__(
        self,
        class_name: str,
        cmd_type: const.CommandType,
        cmd_code: int,
        module_index: int | None,
        port_index: int | None,
        indices: list[int],
        values: RequestBodyStruct | None = None,
    ) -> None:
        self.class_name = class_name
        self.header = ProtocolHeader(
            magic_word=const.MAGIC_WORD,
            number_of_indices=len(indices),
            number_of_value_bytes=values.nbytes() if values else 0,
            cmd_code=cmd_code,
            cmd_type=cmd_type.value,
            module_index=const.NOTHING if module_index is None else module_index,
            port_index=const.NOTHING if port_index is None else port_index,
            request_identifier=NOT_SET_IDENTIFIER,
        )
        self.index_values = indices
        self.values = values

    def __str__(self) -> str:
        return _utils.format_str(self)

    def __repr__(self) -> str:
        return _utils.format_repr(self)

    def __bytes__(self) -> bytes:
        idx_format = const.indices_format(self.header.number_of_indices)
        return bytes().join(
            (
                self.header,
                struct.pack(idx_format, *self.index_values),
                self.values.to_bytes() if self.values else b""
            )
        )

    def update_identifier(self, request_id: int) -> None:
        self.header.request_identifier = request_id

    @property
    def cmd_code(self) -> int:
        return self.header.cmd_code



from typing import NamedTuple


class ModuleKind(NamedTuple):
    """NamedTuple of id's of module."""
    module_id: int


class PortKind(NamedTuple):
    """NamedTuple of id's of module, port."""
    module_id: int
    port_id: int


class IndicesKind(NamedTuple):
    """NamedTuple of id's of module, port, and xindex."""
    module_id: int
    port_id: int
    index_id: int



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.commands import (
    P4G_L2_CLIENT_MAC,
    P4G_L2_SERVER_MAC,
    P4G_L2_USE_ADDRESS_RES,
    P4G_L2_USE_GW,
    P4G_L2_GW,
    P4G_L2_IPV6_GW,
    P4G_VLAN_ENABLE,
    P4G_VLAN_TCI,
)


class GMacL2:
    """L2 MAC"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.client = P4G_L2_CLIENT_MAC(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_L2_CLIENT_MAC
        """
        self.server = P4G_L2_SERVER_MAC(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_L2_SERVER_MAC
        """


class GGatewayL2:
    """L2 Gateway"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.use = P4G_L2_USE_GW(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_L2_USE_GW
        """
        self.ipv4 = P4G_L2_GW(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_L2_GW
        """
        self.ipv6 = P4G_L2_IPV6_GW(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_L2_IPV6_GW
        """


class GVlanL2:
    """L2 VLAN"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.enable = P4G_VLAN_ENABLE(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_VLAN_ENABLE
        """
        self.tci = P4G_VLAN_TCI(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_VLAN_TCI
        """


class GL2:
    """L2 Configurations"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.address_resolve = P4G_L2_USE_ADDRESS_RES(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_L2_USE_ADDRESS_RES
        """
        self.mac = GMacL2(conn, module_id, port_id, group_idx)
        """L2 MAC"""
        self.gateway = GGatewayL2(conn, module_id, port_id, group_idx)
        """L2 Gateway"""
        self.vlan = GVlanL2(conn, module_id, port_id, group_idx)
        """L2 VLAN"""



from __future__ import annotations
import asyncio
from typing import (
    Awaitable,
    Iterable,
    Type,
    TypeVar,
    Protocol,
    TYPE_CHECKING,
    cast,
)
from collections import OrderedDict
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf

from .abc import AbcResourcesManager
from .exceptions import NoSuchPortError
from ..con_traffic_light import ConnectionTrafficLight


class IPort(Protocol):
    _conn: "itf.IConnection"

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        ...


PT = TypeVar("PT", bound=IPort)


class PortBaseManager(AbcResourcesManager[PT]):
    def obtain(self, key: int) -> PT:
        try:
            return super().obtain(key)
        except KeyError:
            raise NoSuchPortError(key)

    def obtain_multiple(self, *keys: int) -> tuple[PT, ...]:
        try:
            return super().obtain_multiple(*keys)
        except KeyError as err:
            raise NoSuchPortError(err.args[0])


class PortsManager(PortBaseManager[PT]):

    __slots__ = ("_conn", "_ports_type", "_ports_count", "_module_id", )

    def __init__(self, conn: "itf.IConnection", module_id: int, ports_type: Type[PT], ports_count: int) -> None:
        super().__init__()
        self._conn = conn
        self._ports_type = ports_type
        self._module_id = module_id
        self._items: OrderedDict[int, PT] = OrderedDict()
        self._ports_count = 0
        self.reinit(ports_count)

    def reinit(self, ports_count: int) -> bool:
        if ports_count == self._ports_count:
            return False
        
        if self._items:
            for v in self._items.values():
                v._conn.set_outdated()
        del self._items
        self._items: OrderedDict[int, PT] = OrderedDict(
            (
                port_id,
                self._ports_type(
                    conn=ConnectionTrafficLight(self._conn),
                    module_id=self._module_id,
                    port_id=port_id,
                )
            )
            for port_id in range(ports_count)
        )
        self._ports_count = ports_count
        return True

    async def fill(self) -> None:
        assert not self._lock, "Method <fill> can be called only once."
        coros = cast(Iterable[Awaitable], self._items.values())
        await asyncio.gather(*coros)


class PortResolver(Protocol):
    async def __call__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> Type:
        ...


class PortsCombiManager(PortBaseManager[PT]):

    __slots__ = ("_conn", "_resolver", "_ports_count", "_module_id",)

    def __init__(self, conn: "itf.IConnection", module_id: int, resolver: PortResolver, ports_count: int) -> None:
        super().__init__()
        self._conn = conn
        self._resolver = resolver
        self._ports_count = ports_count
        self._module_id = module_id
        self._items: OrderedDict[int, PT] = OrderedDict()

    async def fill(self) -> None:
        assert not self._lock, "Method <fill> can be called only once."
        coros = iter(
            self._resolver(
                conn=ConnectionTrafficLight(self._conn),
                module_id=self._module_id,
                port_id=port_id,
            )
            for port_id in range(self._ports_count)
        )
        ports = await asyncio.gather(*coros)
        self._items = OrderedDict(
            (port.kind.port_id, port) for port in ports
        )



from __future__ import annotations
from dataclasses import dataclass
import typing

from xoa_driver.internals.core.builders import (
    build_get_request,
    build_set_request
)
from xoa_driver.internals.core import interfaces
from xoa_driver.internals.core.token import Token
from xoa_driver.internals.core.transporter.registry import register_command
from xoa_driver.internals.core.transporter.protocol.payload import (
    field,
    RequestBodyStruct,
    ResponseBodyStruct,
    XmpByte,
    XmpInt,
    XmpLong,
    XmpSequence,
)
from .enums import OnOff


@register_command
@dataclass
class PR_TPLDJITTER:
    """
    Obtains statistics concerning the jitter experienced by the packets with a
    particular test payload id received on a port. The values are the difference in
    packet-to-packet latency, and the minimum will usually be zero.A special value
    of -1 is returned if jitter numbers are not applicable. They are only available
    for TID values 0..31.
    """

    code: typing.ClassVar[int] = 239
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _test_payload_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        min_val: int = field(XmpLong())
        """long integer, nanoseconds, minimum jitter for test payload stream"""
        avg_val: int = field(XmpLong())
        """long integer, nanoseconds, average jitter for test payload stream"""
        max_val: int = field(XmpLong())
        """long integer, nanoseconds, maximum jitter for test payload stream"""
        avg_last_sec: int = field(XmpLong())
        """long integer, nanoseconds, average jitter over last 1-second period"""
        min_last_sec: int = field(XmpLong())
        """long integer, nanoseconds, minimum jitter during last 1-second period"""
        max_last_sec: int = field(XmpLong())
        """long integer, nanoseconds, maximum jitter during last 1-second period"""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning the jitter experienced by the packets with a particular test payload id received on a port.

        :return: minimum|average|maximum jitter (nanoseconds), average|average|maximum jitter over last 1-second period (nanoseconds)
        :rtype: PR_TPLDJITTER.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._test_payload_xindex]))


@register_command
@dataclass
class PR_TOTAL:
    """
    Obtains statistics concerning all the packets received on a port.
    """

    code: typing.ClassVar[int] = 240
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        bit_count_last_sec: int = field(XmpLong())
        """long integer, number of bits received in the last second."""
        packet_count_last_sec: int = field(XmpLong())
        """long integer, number of packets received in the last second."""
        byte_count_since_cleared: int = field(XmpLong())
        """long integer, number of bytes received since statistics were cleared."""
        packet_count_since_cleared: int = field(XmpLong())
        """long integer, number of packets received since statistics were cleared."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning all the packets received on a port.

        :return:
            number of bits received in the last second,
            number of packets received in the last second,
            number of bytes received since statistics were cleared,
            and number of packets received since statistics were cleared.

        :rtype: PR_TOTAL.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PR_NOTPLD:
    """
    Obtains statistics concerning the packets without a test payload received on a
    port.
    """

    code: typing.ClassVar[int] = 241
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        bit_count_last_sec: int = field(XmpLong())
        """long integer, number of bits received in the last second."""
        packet_count_last_sec: int = field(XmpLong())
        """long integer, number of packets received in the last second."""
        byte_count_since_cleared: int = field(XmpLong())
        """long integer, number of bytes received since statistics were cleared."""
        packet_count_since_cleared: int = field(XmpLong())
        """long integer, number of packets received since statistics were cleared."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning the packets without a test payload received on a port.

        :return:
            number of bits received in the last second,
            number of packets received in the last second,
            number of bytes received since statistics were cleared,
            and number of packets received since statistics were cleared.

        :rtype: PR_NOTPLD.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PR_EXTRA:
    """
    Obtains statistics concerning special errors received on a port since received statistics were cleared.
    """

    code: typing.ClassVar[int] = 242
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        fcs_error_count: int = field(XmpLong())
        """long integer, number of packets with frame checksum errors."""
        pause_frame_count: int = field(XmpLong())
        """long integer, number of Ethernet pause frames."""
        rx_arp_request_count: int = field(XmpLong())
        """long integer, number of ARP request packets received."""
        rx_arp_reply_count: int = field(XmpLong())
        """long integer, number of ARP reply packets received."""
        rx_ping_request_count: int = field(XmpLong())
        """long integer, number of PING request packets received."""
        rx_ping_reply_count: int = field(XmpLong())
        """long integer, number of PING reply packets received."""
        gap_count: int = field(XmpLong())
        """long integer, number of gap-monitored gaps encountered."""
        gap_duration: int = field(XmpLong())
        """long integer, combined duration of gap-monitored gaps encountered, microseconds."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning special packets received on a port since statistics were cleared.

        :return: number of packets with fcs error frames, pause frames, arp rxreq, arp rxrsp, ping rxreq, ping rxrsp, gap events, and gap microseconds.
        :rtype: PR_EXTRA.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PR_TPLDS:
    """
    Obtain the set of test payload IDs observed among the received packets since
    receive statistics were cleared. Traffic statistics for these test payload
    streams will have non-zero byte and packet count.
    """

    code: typing.ClassVar[int] = 243
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        test_payload_identifiers: typing.List[int] = field(XmpSequence(types_chunk=[XmpInt()]))
        """list of integers, the identifiers of the test payload."""

    def get(self) -> Token[GetDataAttr]:
        """Get the set of test payload IDs observed among the received packets since
        receive statistics were cleared. Traffic statistics for these test payload
        streams will have non-zero byte and packet count.

        :return: the identifiers of the test payload
        :rtype: PR_TPLDS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PR_TPLDTRAFFIC:
    """
    Obtains traffic statistics concerning the packets with a particular test payload
    identifier received on a port.
    """

    code: typing.ClassVar[int] = 244
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _test_payload_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        bit_count_last_sec: int = field(XmpLong())
        """long integer, number of bits received in the last second."""
        packet_count_last_sec: int = field(XmpLong())
        """long integer, number of packets received in the last second."""
        byte_count_since_cleared: int = field(XmpLong())
        """long integer, number of bytes received since statistics were cleared."""
        packet_count_since_cleared: int = field(XmpLong())
        """long integer, number of packets received since statistics were cleared."""

    def get(self) -> Token[GetDataAttr]:
        """Get traffic statistics concerning the packets with a particular test payload
        identifier received on a port.

        :return:
            number of bits received in the last second,
            number of packets received in the last second,
            number of bytes received since statistics were cleared,
            number of packets received since statistics were cleared

        :rtype: PR_TPLDTRAFFIC.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._test_payload_xindex]))


@register_command
@dataclass
class PR_TPLDERRORS:
    """
    Obtains statistics concerning errors in the packets with a particular test
    payload id received on a port. The error information is derived from analyzing
    the various fields contained in the embedded test payloads of the received
    packets, independent of which chassis and port may have originated the packets.
    Note that packet-lost statistics involve both a transmitting port and a
    receiving port, and in particular knowing which port originated the packets with
    a particular test payload identifier. This information requires knowledge of the
    global test environment, and is not supported at the port-level.
    """

    code: typing.ClassVar[int] = 245
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _test_payload_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        dummy: int = field(XmpLong())
        """long integer, not in use."""
        non_incre_seq_event_count: int = field(XmpLong())
        """long integer, number of non-incrementing-sequence-number events."""
        swapped_seq_misorder_event_count: int = field(XmpLong())
        """long integer, number of swapped-sequence-number misorder events."""
        non_incre_payload_packet_count: int = field(XmpLong())
        """long integer, number of packets with non-incrementing payload content."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning errors in the packets with a particular test payload id received on a port.

        :return:
            dummy value not in use,
            number of non-incrementing-sequence-number events,
            number of swapped-sequence-number misorder events,
            number of packets with non-incrementing payload content

        :rtype: PR_TPLDERRORS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._test_payload_xindex]))


@register_command
@dataclass
class PR_TPLDLATENCY:
    """
    Obtains statistics concerning the latency experienced by the packets with a
    particular test payload id received on a port. The values are adjusted by the
    port-level P_LATENCYOFFSET value. A special value of -1 is returned if latency
    numbers are not applicable. Latency is only meaningful when the clocks of the
    transmitter and receiver are synchronized. This requires the two ports to be on
    the same test module, and it requires knowledge of the global test environment
    to ensure that packets are in fact routed between these ports.
    """

    code: typing.ClassVar[int] = 246
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _test_payload_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        min_val: int = field(XmpLong())
        """long integer, nanoseconds, minimum latency for test payload stream"""
        avg_val: int = field(XmpLong())
        """long integer, nanoseconds, average latency for test payload stream"""
        max_val: int = field(XmpLong())
        """long integer, nanoseconds, maximum latency for test payload stream"""
        avg_last_sec: int = field(XmpLong())
        """long integer, nanoseconds, average latency over last 1-second period"""
        min_last_sec: int = field(XmpLong())
        """long integer, nanoseconds, minimum latency during last 1-second period"""
        max_last_sec: int = field(XmpLong())
        """long integer, nanoseconds, maximum latency during last 1-second period"""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning the latency experienced by the packets with a particular test payload id received on a port.

        :return: minimum|average|maximum latency (nanoseconds), average|average|maximum latency over last 1-second period (nanoseconds)
        :rtype: PR_TPLDLATENCY.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._test_payload_xindex]))


@register_command
@dataclass
class PR_FILTER:
    """
    Obtains statistics concerning the packets satisfying the condition of a
    particular filter for a port.
    """

    code: typing.ClassVar[int] = 247
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _filter_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        bit_count_last_sec: int = field(XmpLong())
        """long integer, number of bits received in the last second."""
        packet_count_last_sec: int = field(XmpLong())
        """long integer, number of packets received in the last second."""
        byte_count_since_cleared: int = field(XmpLong())
        """long integer, number of bytes received since statistics were cleared."""
        packet_count_since_cleared: int = field(XmpLong())
        """long integer, number of packets received since statistics were cleared."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning the packets satisfying the condition of a particular filter for a port

        :return:
            number of bits received in the last second,
            number of packets received in the last second,
            number of bytes received since statistics were cleared,
            number of packets received since statistics were cleared

        :rtype: PR_FILTER.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._filter_xindex]))


@register_command
@dataclass
class PR_CLEAR:
    """
    Clear all the receive statistics for a port. The byte and packet counts will
    restart at zero.
    """

    code: typing.ClassVar[int] = 248
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class SetDataAttr(RequestBodyStruct):
        pass

    def set(self) -> Token[None]:
        """Clear all the receive statistics for a port. The byte and packet counts will
        restart at zero.
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PR_CALIBRATE:
    """
    Calibrate the latency calculation for packets received on a port. The lowest
    detected latency value (across all Test Payload IDs) will be set as the new
    base.
    """

    code: typing.ClassVar[int] = 249
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class SetDataAttr(RequestBodyStruct):
        pass

    def set(self) -> Token[None]:
        """Calibrate the latency calculation for packets received on a port. The lowest
        detected latency value (across all Test Payload IDs) will be set as the new
        base.
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PR_UAT_STATUS:
    """
    This command will show the current UAT (UnAvailable Time) state, which is used
    in Valkyrie1564
    """

    code: typing.ClassVar[int] = 252
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        status: OnOff = field(XmpByte())
        """coded byte, specifies the state of the affected stream counter."""

    def get(self) -> Token[GetDataAttr]:
        """Get the current UAT (UnAvailable Time) state, which is used
        in Valkyrie1564.

        :return: the state of the affected stream counter
        :rtype: PR_UAT_STATUS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PR_UAT_TIME:
    """
    This command will show the current number of unavailable seconds, which is used in Valkyrie1564.
    """

    code: typing.ClassVar[int] = 256
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        time: int = field(XmpInt())
        """integer, number of unavailable seconds."""

    def get(self) -> Token[GetDataAttr]:
        """Get the current UAT (UnAvailable Time) state, which is used
        in Valkyrie1564.

        :return: number of unavailable seconds
        :rtype: PR_UAT_TIME.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PR_TOTALEXT:
    """
    .. versionadded: v1.1

    An extension of :class:`PR_TOTAL` that also includes a calculation of bytes received in the last second, as well as a number of port error counters.
    PR_TOTALEXT returns list of long integers. This list may be expanded in future software releases.
    """

    code: typing.ClassVar[int] = 257
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        bit_count_last_sec: int = field(XmpLong())
        """long integer, number of bits received in the last second."""
        byte_count_last_sec: int = field(XmpLong())
        """long integer, number of bytes received in the last second."""
        packet_count_last_sec: int = field(XmpLong())
        """long integer, number of packets received in the last second."""
        byte_count_since_cleared: int = field(XmpLong())
        """long integer, number of bytes received since statistics were cleared."""
        packet_count_since_cleared: int = field(XmpLong())
        """long integer, number of packets received since statistics were cleared."""
        fcs_error_count: int = field(XmpLong())

        oversize_count: int = field(XmpLong())

        undersize_count: int = field(XmpLong())

        jabber_count: int = field(XmpLong())

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning all the packets received on a port.

        :return:
            number of bits received in the last second,
            number of bytes received in the last second,
            number of packets received in the last second,
            number of bytes received since statistics were cleared,
            number of packets received since statistics were cleared,
            number of packets received with fcs error frames,
            number of oversize packets received since last clear (-1 if this counter is not supported by the tester),
            number of undersize packets received since last clear (-1 if this counter is not supported by the tester),
            number of jabber packets received since last clear (-1 if this counter is not supported by the tester).

        :rtype: PR_TOTALEXT.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PR_NOTPLDEXT:
    """
    .. versionadded: v1.1

    An extension of :class:`PR_NOTPLD` that also includes a calculation of bytes received in the last second.
    PR_NOTPLDEXT returns list of long integers. This list may be expanded in future software releases.
    """

    code: typing.ClassVar[int] = 258
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        bit_count_last_sec: int = field(XmpLong())
        """long integer, number of bits received in the last second."""
        byte_count_last_sec: int = field(XmpLong())
        """long integer, number of bytes received in the last second."""
        packet_count_last_sec: int = field(XmpLong())
        """long integer, number of packets received in the last second."""
        byte_count_since_cleared: int = field(XmpLong())
        """long integer, number of bytes received since statistics were cleared."""
        packet_count_since_cleared: int = field(XmpLong())
        """long integer, number of packets received since statistics were cleared."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning the packets without a test payload received on a port.

        :return:
            number of bits received in the last second,
            number of bytes received in the last second,
            number of packets received in the last second,
            number of bytes received since statistics were cleared,
            and number of packets received since statistics were cleared.

        :rtype: PR_NOTPLDEXT.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PR_TPLDTRAFFICEXT:
    """
    .. versionadded: v1.1

    An extension of :class:`PR_TPLDTRAFFIC` that also includes a calculation of bytes received in the last second.
    PR_TPLDTRAFFICEXT returns list of long integers. This list may be expanded in future software releases.
    """

    code: typing.ClassVar[int] = 259
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _test_payload_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        bit_count_last_sec: int = field(XmpLong())
        """long integer, number of bits received in the last second."""
        byte_count_last_sec: int = field(XmpLong())
        """long integer, number of bytes received in the last second."""
        packet_count_last_sec: int = field(XmpLong())
        """long integer, number of packets received in the last second."""
        byte_count_since_cleared: int = field(XmpLong())
        """long integer, number of bytes received since statistics were cleared."""
        packet_count_since_cleared: int = field(XmpLong())
        """long integer, number of packets received since statistics were cleared."""

    def get(self) -> Token[GetDataAttr]:
        """Get traffic statistics concerning the packets with a particular test payload
        identifier received on a port.

        :return:
            number of bits received in the last second,
            number of bytes received in the last second,
            number of packets received in the last second,
            number of bytes received since statistics were cleared,
            number of packets received since statistics were cleared

        :rtype: PR_TPLDTRAFFICEXT.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._test_payload_xindex]))


@register_command
@dataclass
class PR_FILTEREXT:
    """
    .. versionadded: v1.1

    An extension of :class:`PR_FILTER` that also includes a calculation of bytes received in the last second.
    PR_FILTEREXT returns list of long integers. This list may be expanded in future software releases.
    """

    code: typing.ClassVar[int] = 260
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _filter_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        bit_count_last_sec: int = field(XmpLong())
        """long integer, number of bits received in the last second."""
        byte_count_last_sec: int = field(XmpLong())
        """long integer, number of bytes received in the last second."""
        packet_count_last_sec: int = field(XmpLong())
        """long integer, number of packets received in the last second."""
        byte_count_since_cleared: int = field(XmpLong())
        """long integer, number of bytes received since statistics were cleared."""
        packet_count_since_cleared: int = field(XmpLong())
        """long integer, number of packets received since statistics were cleared."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning the packets satisfying the condition of a particular filter for a port

        :return:
            number of bits received in the last second,
            number of bytes received in the last second,
            number of packets received in the last second,
            number of bytes received since statistics were cleared,
            number of packets received since statistics were cleared

        :rtype: PR_FILTEREXT.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._filter_xindex]))


@register_command
@dataclass
class PR_PFCSTATS:
    """
    Obtains statistics of received Priority Flow Control (PFC) packets on a port.

    .. versionchanged:: 1.1
    """

    code: typing.ClassVar[int] = 374
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        packet_count: int = field(XmpLong())
        """long integer, the total number of Priority Flow Control (PFC) packets received since statistics were cleared."""
        quanta_pri_0: int = field(XmpLong())
        """the total number of valid PFC quanta received on the port for priority level 0 since statistics were cleared"""
        quanta_pri_1: int = field(XmpLong())
        """the total number of valid PFC quanta received on the port for priority level 1 since statistics were cleared"""
        quanta_pri_2: int = field(XmpLong())
        """the total number of valid PFC quanta received on the port for priority level 2 since statistics were cleared"""
        quanta_pri_3: int = field(XmpLong())
        """the total number of valid PFC quanta received on the port for priority level 3 since statistics were cleared"""
        quanta_pri_4: int = field(XmpLong())
        """the total number of valid PFC quanta received on the port for priority level 4 since statistics were cleared"""
        quanta_pri_5: int = field(XmpLong())
        """the total number of valid PFC quanta received on the port for priority level 5 since statistics were cleared"""
        quanta_pri_6: int = field(XmpLong())
        """the total number of valid PFC quanta received on the port for priority level 6 since statistics were cleared"""
        quanta_pri_7: int = field(XmpLong())
        """the total number of valid PFC quanta received on the port for priority level 7 since statistics were cleared"""

    def get(self) -> Token[GetDataAttr]:
        """Get the statistics of received Priority Flow Control (PFC) packets on a port.

        :return: the total number of Priority Flow Control (PFC) packets received since statistics were cleared
        :rtype: PR_PFCSTATS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PR_FLOWTOTAL:
    """
    Obtains statistics concerning all the packets received from a flow between this
    receive port and its partner TX port.
    """

    code: typing.ClassVar[int] = 1741
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        l2_bps: int = field(XmpLong())
        """long integer, number of bits received at layer 2 in the last second for the flow."""
        pps: int = field(XmpLong())
        """long integer, number of packets received in the last second for the flow."""
        byte_count: int = field(XmpLong())
        """long integer, number of bytes received since statistics were cleared for the flow."""
        packet_count: int = field(XmpLong())
        """long integer, number of packets received since statistics were cleared for the flow."""

    def get(self) -> Token[GetDataAttr]:
        """(Chimera only) Get statistics concerning all the packets received from a flow between this receive port and its partner TX port.

        :return:
            number of bits received at layer 2 in the last second for the flow,
            number of packets received in the last second for the flow,
            number of bytes received since statistics were cleared for the flow,
            number of packets received since statistics were cleared for the flow

        :rtype: PR_FLOWTOTAL.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))


@register_command
@dataclass
class PR_FLOWCLEAR:
    """
    Clear all the receive statistics on a particular flow for a Chimera port. The
    byte and packet counts will restart at zero.
    """

    code: typing.ClassVar[int] = 1743
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class SetDataAttr(RequestBodyStruct):
        pass

    def set(self) -> Token[None]:
        """(Chimera only) Clear all the receive statistics on a particular flow for a Chimera port. The byte and packet counts will restart at zero.
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))



import asyncio
import functools
import typing
from typing_extensions import Self
from xoa_driver.internals.commands import (
    M_LICENSE_DEMO_INFO,
    M_LICENSE_MAINTENANCE_INFO,
    M_LICENSE_CWB_DETECTED,
    M_LICENSE_UPDATE,
    M_LICENSE_UPDATE_STATUS,
    M_LICENSE_LIST_BSON,
    M_LICENSE_ONLINE,
    # M4_INT, #TODO: need to implement
    # M4_STRING, #TODO: need to implement
    # M4_HEX, #TODO: need to implement
    M4_SYSTEMID,
    M4_VERSIONNO,
    M4_SYSTEM_STATUS,
    M4_COMPATIBLE_CLIENT_VERSION,
    M4_TIME,
    M4_SYSTEM_TIME,
    M4_MEM_INFO,
    M4_CAPTURE_SIZE,
    M4_LICENSE_INFO,
    M4_REPLAY_PARSE_START,
    M4_REPLAY_PARSE_STOP,
    M4_REPLAY_PARSE_STATE,
    M4_REPLAY_PARSER_PARAMS,
    M4_REPLAY_FILE_LIST_BSON,
    M4_REPLAY_FILE_LIST,
    M4_CAPTURE_FILE_LIST_BSON,
    M4_CAPTURE_FILE_LIST,
    # M4_REPLAY_FILE_INFO_BSON, #TODO: need to implement
    M4_REPLAY_FILE_DELETE,
    M4_CAPTURE_FILE_DELETE,
    M4E_MODE,
    M4E_RESERVE,
    M4_TLS_CIPHER_SUITES,
)
from xoa_driver.internals.hli_v1 import revisions
from xoa_driver.internals.utils import attributes as utils
from xoa_driver.internals.utils.managers import ports_manager as pm
from xoa_driver.internals.state_storage import modules_state
from xoa_driver import ports
from . import base_module as bm
if typing.TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from . import __interfaces as m_itf


class License:
    """
    License management of a Vulcan tester.
    """

    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.demo_info = M_LICENSE_DEMO_INFO(conn, module_id)
        """
        Returns info about the demo status of the module.

        :type: M_LICENSE_DEMO_INFO
        """

        self.management_info = M_LICENSE_MAINTENANCE_INFO(conn, module_id)
        """
        Returns info about the maintenance license status for the module.

        :type: M_LICENSE_MAINTENANCE_INFO
        """

        self.clock_windback = M_LICENSE_CWB_DETECTED(conn, module_id)
        """
        Returns if clock-windback is detected. If clock-windback has been detected the chassis is locked and no reservations of ports can be performed.

        :type: M_LICENSE_CWB_DETECTED
        """

        self.update = M_LICENSE_UPDATE(conn, module_id)
        """
        Instructs the chassis to update its local license information.

        :type: M_LICENSE_UPDATE
        """

        self.update_status = M_LICENSE_UPDATE_STATUS(conn, module_id)
        """
        Returns the status of the latest license update operations.

        :type: M_LICENSE_UPDATE_STATUS
        """

        self.list_bson = M_LICENSE_LIST_BSON(conn, module_id)
        """
        Returns a list of locally stored licenses - formatted as a BSON document.

        :type: M_LICENSE_LIST_BSON
        """

        self.online_mode = M_LICENSE_ONLINE(conn, module_id)
        """
        Configures the chassis in online or offline mode.

        :type: M_LICENSE_ONLINE
        """


class CaptureParse:
    """
    Parse a captured pcap file.
    """

    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.start = M4_REPLAY_PARSE_START(conn, module_id)
        """
        To start parsing an uploaded Capture File (in PCAP format) intended for use in a replay test scenario.

        :type: M4_REPLAY_PARSE_START
        """

        self.stop = M4_REPLAY_PARSE_STOP(conn, module_id)
        """
        To stop parsing an uploaded Capture File (in PCAP format) intended for use in a replay test scenario.

        :type: M4_REPLAY_PARSE_STOP
        """

        self.state = M4_REPLAY_PARSE_STATE(conn, module_id)
        """
        Returns the state of the parser.

        :type: M4_REPLAY_PARSE_STATE
        """

        self.parser_params = M4_REPLAY_PARSER_PARAMS(conn, module_id)
        """
        Configuration of parameters for the parsing of pcap files.

        :type: M4_REPLAY_PARSER_PARAMS
        """


class ReplayFile:
    """
    The pcap files to replay.
    """

    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.list_bson = M4_REPLAY_FILE_LIST_BSON(conn, module_id)
        """
        Returns a list of replay files formatted in BSON.

        :type: M4_REPLAY_FILE_LIST_BSON
        """
        self.list = M4_REPLAY_FILE_LIST(conn, module_id)
        """
        Returns a list of replay files.

        :type: M4_REPLAY_FILE_LIST
        """
        # self.info_bson = M4_REPLAY_FILE_INFO_BSON(conn, module_id)  # TODO: need to implement

        self.delete = M4_REPLAY_FILE_DELETE(conn, module_id)
        """
        Delete a replay file.

        :type: M4_REPLAY_FILE_DELETE
        """


class Replay:
    """
    PCAP replay settings and control.
    """

    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.file = ReplayFile(conn, module_id)
        """
        Replay file settings and control.

        :type: ReplayFile
        """


class Capture:
    """
    Captured pcap file settings and control.
    """

    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.size = M4_CAPTURE_SIZE(conn, module_id)
        """
        Specify whether to capture whole packets(large) or truncated packets.

        :type: M4_CAPTURE_SIZE
        """

        self.file_list_bson = M4_CAPTURE_FILE_LIST_BSON(conn, module_id)
        """
        Returns a list of captured files formatted as a BSON.

        :type: M4_CAPTURE_FILE_LIST_BSON
        """

        self.file_list = M4_CAPTURE_FILE_LIST(conn, module_id)
        """
        Returns a list of captured files.

        :type: M4_CAPTURE_FILE_LIST
        """

        self.file_delete = M4_CAPTURE_FILE_DELETE(conn, module_id)
        """
        Delete a captured file.

        :type: M4_CAPTURE_FILE_DELETE
        """

        self.parse = CaptureParse(conn, module_id)
        """
        Capture file parser settings.

        :type: CaptureParse
        """


class PacketEngine:
    """
    Packet engine.
    """

    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.license_info = M4_LICENSE_INFO(conn, module_id)
        """
        Returns the number of available and free PE licenses.

        :type: M4_LICENSE_INFO
        """

        self.reserve = M4E_RESERVE(conn, module_id)
        """
        Reserve a number of PEs so they later can be assigned to specific ports.

        :type: M4E_RESERVE
        """

        self.mode = M4E_MODE(conn, module_id)
        """
        Select resource allocation mode.

        :type: M4E_MODE
        """


class ModuleSystem:
    """
    L47 module info.
    """

    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.id = M4_SYSTEMID(conn, module_id)
        """
        The system identifier of the module.

        :type: M4_SYSTEMID
        """

        self.status = M4_SYSTEM_STATUS(conn, module_id)
        """
        The module system status in a text string.

        :type: M4_SYSTEM_STATUS
        """

        self.time = M4_SYSTEM_TIME(conn, module_id)
        """
        The module system time in UTC.
        :type: M4_SYSTEM_TIME
        """


@revisions.register_vulcan_module(rev="Xena L47 Module")
class ModuleL47(bm.BaseModule["modules_state.ModuleLocalState"]):
    """
    This is a conceptual class of L47 test module on a Vulcan tester.
    """

    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)

        self._local_states = modules_state.ModuleLocalState()

        self.version_number = M4_VERSIONNO(conn, self.module_id)
        """
        Version string containing a combination of information regarding the software version and the build environment. The first part of the string is the software build version.

        :type: M4_VERSIONNO
        """

        self.compatible_client_version = M4_COMPATIBLE_CLIENT_VERSION(conn, self.module_id)
        """
        The recommended and required VulcanMananger client version.

        :type: M4_COMPATIBLE_CLIENT_VERSION
        """

        self.time = M4_TIME(conn, self.module_id)
        """
        The module time in millisecond.

        :type: M4_TIME
        """

        self.tls_cipher = M4_TLS_CIPHER_SUITES(conn, self.module_id)
        """
        The list of supported TLS Cipher Suites.

        :type: M4_TLS_CIPHER_SUITES
        """

        self.memory_info = M4_MEM_INFO(conn, self.module_id)  # Not sure if suppose to be here, maybe under ModuleSystem
        """
        The system memory information.

        :type: M4_MEM_INFO
        """

        self.module_system = ModuleSystem(conn, self.module_id)
        """L47 module info.

        :type: ModuleSystem
        """

        self.license = License(conn, self.module_id)
        """L47 license management.

        :type: License
        """

        self.replay = Replay(conn, self.module_id)
        """L47 pcap replay.

        :type: Replay
        """

        self.capture = Capture(conn, self.module_id)
        """L47 capture to pcap files.

        :type: Capture
        """

        self.packet_engine = PacketEngine(conn, self.module_id)
        """L47 packet engine.

        :type: ~xoa_driver.internals.hli_v1.modules.module_l47.PacketEngine
        """

        self.ports: pm.PortsManager[ports.PortL47] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PortL47,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """
        L47 Port index manager of this test module.

        :type: PortsManager
        """

    @property
    def info(self) -> modules_state.ModuleLocalState:
        """Return the module's local state

        :return: the module's local state
        :rtype: ModuleLocalState
        """

        return self._local_states

    async def _setup(self) -> Self:
        await asyncio.gather(
            self._local_states.initiate(self),
            self.ports.fill()
        )
        self._local_states.register_subscriptions(self)
        return self

    on_license_demo_info_change = functools.partialmethod(utils.on_event, M_LICENSE_DEMO_INFO)
    """Register a callback to the event that the module's demo license info changes."""

    on_license_maintenance_info_change = functools.partialmethod(utils.on_event, M_LICENSE_MAINTENANCE_INFO)
    """Register a callback to the event that the module's maintenance license info changes."""

    on_replay_parse_state_change = functools.partialmethod(utils.on_event, M4_REPLAY_PARSE_STATE)
    """Register a callback to the event that the module's parsing state of a captured pcap file changes."""

    on_license_info_change = functools.partialmethod(utils.on_event, M4_LICENSE_INFO)
    """Register a callback to the event that the module's L47 license info changes."""



from __future__ import annotations
from dataclasses import dataclass
import typing
import functools

from xoa_driver.internals.core.builders import (
    build_get_request,
    build_set_request
)
from xoa_driver.internals.core import interfaces
from xoa_driver.internals.core.token import Token
from xoa_driver.internals.core.transporter.registry import register_command
from xoa_driver.internals.core.transporter.protocol.payload import (
    field,
    RequestBodyStruct,
    ResponseBodyStruct,
    XmpHex,
    XmpInt,
    XmpLong,
    XmpSequence,
    Hex,
)
from .enums import (
    StartTrigger,
    StopTrigger,
    PacketType,
)


@register_command
@dataclass
class PC_TRIGGER:
    """
    The criteria for when to start and stop the capture process for a port. Even
    when capture is enabled with :class:`P_CAPTURE`, the actual capturing of packets can be
    delayed until a particular start criteria is met by a received packet.
    Likewise, a stop criteria can be specified, based on a received packet. If no
    explicit stop criteria is specified, capture  stops when the internal buffer
    runs full. In buffer overflow situations, if there is an explicit  stop
    criteria, then the latest packets will be retained (and the early ones
    discarded),  and otherwise, the earliest packets are retained (and the later
    ones discarded).
    """

    code: typing.ClassVar[int] = 221
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        start_criteria: StartTrigger = field(XmpInt())
        """The criteria for starting the actual packet capture.

        :type: StartTrigger
        """
        start_criteria_filter: int = field(XmpInt())
        """The index of a particular filter for the start criteria.

        :type: int
        """
        stop_criteria: StopTrigger = field(XmpInt())
        """The criteria for stopping the actual packet capture.

        :type: StopTrigger
        """
        stop_criteria_filter: int = field(XmpInt())
        """The index of a particular filter for the stop criteria.

        :type: int
        """

    class SetDataAttr(RequestBodyStruct):
        start_criteria: StartTrigger = field(XmpInt())
        """coded integer, the criteria for starting the actual packet capture"""
        start_criteria_filter: int = field(XmpInt())
        """integer, the index of a particular filter for the start criteria."""
        stop_criteria: StopTrigger = field(XmpInt())
        """coded integer, the criteria for stopping the actual packet capture"""
        stop_criteria_filter: int = field(XmpInt())
        """integer, the index of a particular filter for the stop criteria."""

    def get(self) -> Token[GetDataAttr]:
        """Get the capture criteria configurations.

        :return: capture criteria configuration.
        :rtype: ~PC_TRIGGER.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, start_criteria: StartTrigger, start_criteria_filter: int, stop_criteria: StopTrigger, stop_criteria_filter: int) -> Token[None]:
        """Set the capture criteria configurations.

        :param start_criteria: the criteria for starting the actual packet capture
        :type start_criteria: StartTrigger
        :param start_criteria_filter: the index of a particular filter for the start criteria
        :type start_criteria_filter: int
        :param stop_criteria: the criteria for stopping the actual packet capture
        :type stop_criteria: StopTrigger
        :param stop_criteria_filter: the index of a particular filter for the stop criteria
        :type stop_criteria_filter: int
        """

        return Token(
            self._connection,
            build_set_request(
                self,
                module=self._module,
                port=self._port,
                start_criteria=start_criteria,
                start_criteria_filter=start_criteria_filter,
                stop_criteria=stop_criteria,
                stop_criteria_filter=stop_criteria_filter
            )
        )


@register_command
@dataclass
class PC_KEEP:
    """
    Which packets to keep once the start criteria has been triggered for a port.
    Also how big a portion of each packet to retain, saving space for more packets
    in the capture buffer.
    """

    code: typing.ClassVar[int] = 222
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        kind: PacketType = field(XmpInt())
        """The type of packet to keep.

        :type: PacketType
        """
        index: int = field(XmpInt())
        """Test payload id or filter index for which packets to keep.

        :type: int
        """
        byte_count: int = field(XmpInt())
        """How many bytes to keep in the buffer for of each packet. The value -1 means no limit on packet size.

        :type: int
        """

    class SetDataAttr(RequestBodyStruct):
        kind: PacketType = field(XmpInt())
        """coded integer, which general kind of packets to keep"""
        index: int = field(XmpInt())
        """integer, test payload id or filter index for which packets to keep."""
        byte_count: int = field(XmpInt())
        """integer, how many bytes to keep in the buffer for of each packet. The value -1 means no limit on packet size."""

    def get(self) -> Token[GetDataAttr]:
        """Get the configuration of how to keep captured packets.

        :return: the configuration of how to keep captured packets
        :rtype: ~PC_KEEP.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, kind: PacketType, index: int, byte_count: int) -> Token[None]:
        """Set the configuration of how to keep captured packets.

        :param kind: which general kind of packets to keep
        :type kind: PacketType
        :param index: test payload id or filter index for which packets to keep
        :type index: int
        :param byte_count: how many bytes to keep in the buffer for of each packet. The value -1 means no limit on packet size.
        :type byte_count: int
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, kind=kind, index=index, byte_count=byte_count))

    set_all = functools.partialmethod(set, PacketType.ALL)
    """Keep all types of packets.
    """

    set_fcserr = functools.partialmethod(set, PacketType.FCSERR)
    """Only keep the packets with FCS errors.
    """

    set_notpld = functools.partialmethod(set, PacketType.NOTPLD)
    """Only keep the packets without test payload.
    """

    set_tpld = functools.partialmethod(set, PacketType.TPLD)
    """Only keep the packets with test payload.
    """

    set_filter = functools.partialmethod(set, PacketType.FILTER)
    """Only keep the packets that satisfy the given filter.
    """

    set_plderr = functools.partialmethod(set, PacketType.PLDERR)
    """Only keep the packets with payload error in.
    """


@register_command
@dataclass
class PC_STATS:
    """
    Obtains the number of packets currently in the capture buffer for a port. The
    count is reset to zero when capture is turned on.
    """

    code: typing.ClassVar[int] = 224
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        status: int = field(XmpLong())
        """Capture status, 1 if capture has been stopped because of overflow, 0 if still running.

        :type: int
        """
        packets: int = field(XmpLong())
        """The number of packets in the buffer.

        :type: int
        """
        start_time: int = field(XmpLong())
        """Time when capture was started, in **nanoseconds** since 2010-01-01.

        :type: int
        """

    def get(self) -> Token[GetDataAttr]:
        """Get the number of packets currently in the capture buffer for a port. The count is reset to zero when capture is turned on.

        :return: status of the capture, number of packets in the buffer, and start time in nanoseconds since 2010-01-01.
        :rtype: ~PC_STATS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PC_EXTRA:
    """
    Obtains extra information about a captured packet for a port. The information
    comprises time of capture, latency, inter-frame gap, and original packet length.
    Latency is only valid for packets with test payloads and where the originating
    port is on the same module and therefore has the same clock.
    """

    code: typing.ClassVar[int] = 225
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _capture_packet_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        time_captured: int = field(XmpLong())
        """
        Time of capture, in **nanoseconds** since 2010-01-01.

        :type: int
        """
        latency: int = field(XmpLong())
        """
        The number of **nanoseconds** since the packet was transmitted.

        :type: int
        """
        byte_time_count: int = field(XmpLong())
        """
        The number of byte-times since previous packet.

        :type: int
        """
        length: int = field(XmpInt())
        """
        The real length of the packet on the wire.

        :type: int
        """

    def get(self) -> Token[GetDataAttr]:
        """Get extra information about a captured packet for a port.

        :return: Extra information about a captured packet for a port.

        :rtype: ~PC_EXTRA.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._capture_packet_xindex]))


@register_command
@dataclass
class PC_PACKET:
    """
    Obtains the raw bytes of a captured packet for a port. The packet data may be
    truncated if the :class:`PC_KEEP` command specified a limit on the number of bytes kept.
    """

    code: typing.ClassVar[int] = 226
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _capture_packet_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        hex_data: Hex = field(XmpHex())
        """The raw bytes kept for the packet.

        :type: list
        """

    def get(self) -> Token[GetDataAttr]:
        """Get the raw bytes of a captured packet for a port.

        :return: the raw bytes of a captured packet
        :rtype: ~PC_PACKET.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._capture_packet_xindex]))



from __future__ import annotations
from typing import (
    TypeVar,
    Awaitable,
    Generic,
    Callable,
    Generator,
    Any
)
from abc import (
    ABC,
    abstractmethod,
)
import functools
from xoa_driver.internals.commands import enums
from xoa_driver.internals.commands import (
    C_RESERVATION,
    C_DOWN,
    C_PASSWORD,
    C_TIME,
    C_CAPABILITIES,
    C_DEBUGLOGS,
    C_NAME,
    C_MODEL,
    C_COMMENT,
    C_VERSIONNO,
    C_SERIALNO,
    C_RESERVEDBY,
)
from xoa_driver.internals.core.funcs import establish_connection
from xoa_driver.internals.core.transporter.handler import TransportationHandler
from xoa_driver.internals.core.transporter.logger import CustomLogger

from xoa_driver.internals.utils import session
from xoa_driver.internals.state_storage import testers_state


T = TypeVar('T', bound="BaseTester")
TesterStateStorage = TypeVar('TesterStateStorage', bound="testers_state.TesterLocalState")


# min version = rel v83.2 eq fw v446.5
class BaseTester(ABC, Generic[TesterStateStorage]):
    """
    Basic Tester class of APIv2.

    :param host: tester's address/hostname
    :type host: str
    :param username: username of the user
    :type username: str
    :param password: login password of the tester, defaults to "xena"
    :type password: str, optional
    :param port: the port number for connection establishment, default to 22606
    :type port: int, optional
    :param enable_logging: `True` if debug log output from the tester is needed, and `False` otherwise
    :type enable_logging: bool
    :param custom_logger: `None` allow to specify logger with custom configurations
    :type enable_logging: CustomLogger | None
    :param session_timeout: `130` Session timeout in seconds
    :type session_timeout: int
    """

    def __init__(self, host: str, username: str, password: str = "xena", port: int = 22606, *, enable_logging: bool = False, custom_logger: CustomLogger | None = None, session_timeout: int = 130) -> None:
        self.__host = host
        self.__port = port
        self._conn = TransportationHandler(
            enable_logging=enable_logging,
            custom_logger=custom_logger
        )
        self.session = session.TesterSession(
            self._conn,
            username,
            password=password,
            timeout_seconds=session_timeout,
            keepalive=True,
        )
        """
        Current management session
        """

        self.name = C_NAME(self._conn)
        """
        Tester's name.
        Representation of C_NAME
        """

        self.comment = C_COMMENT(self._conn)
        """Description of the tester.
        Representation of C_COMMENT
        """

        self.model = C_MODEL(self._conn)
        """Tester's model.
        Representation of C_MODEL
        """

        self.version_no = C_VERSIONNO(self._conn)
        """Tester's version number.
        Representation of C_VERSIONNO
        """

        self.serial_no = C_SERIALNO(self._conn)
        """Tester's serial number.
        Representation of C_SERIALNO
        """

        self.reservation = C_RESERVATION(self._conn)
        """Tester's reservation operation.
        Representation of C_RESERVATION
        """

        self.reserved_by = C_RESERVEDBY(self._conn)
        """Tester's reservation status.
        Representation of C_RESERVEDBY
        """

        self.down = C_DOWN(self._conn)
        """Shut down the tester.
        Representation of C_DOWN
        """

        self.password = C_PASSWORD(self._conn)
        """Tester's password.
        Representation of C_PASSWORD
        """

        self.time = C_TIME(self._conn)
        """Tester's time in seconds.
        Representation of C_TIME
        """

        self.capabilities = C_CAPABILITIES(self._conn)
        """Tester's capabilities.
        Representation of C_CAPABILITIES
        """

        self.debug_log = C_DEBUGLOGS(self._conn)
        """Tester's debug log.
        Representation of C_DEBUGLOGS
        """

    async def __aenter__(self: Awaitable[T]) -> T:
        return await self

    async def __aexit__(self, type, value, traceback) -> None:
        await self.session.logoff()

    def __await__(self: T) -> Generator[Any, None, T]:
        if self.session.is_online:
            async def skip():
                return self
            return skip().__await__()
        return self._setup().__await__()

    async def _setup(self: T) -> T:
        await establish_connection(self._conn, self.__host, self.__port)
        await self.session.logon()
        return self

    def __is_reservation(self, reserved_status: enums.ReservedStatus) -> bool:
        return self.info.reservation == reserved_status

    is_released = functools.partialmethod(__is_reservation, enums.ReservedStatus.RELEASED)
    """Validate if the tester is released.
    """

    is_reserved_by_me = functools.partialmethod(__is_reservation, enums.ReservedStatus.RESERVED_BY_YOU)
    """Validate if the tester is reserved by my connection.
    """

    @property
    @abstractmethod
    def info(self) -> TesterStateStorage:
        """
        Tester's local information.
        """
        raise NotImplementedError()

    # region Events

    # We are not supporting Subscription on Connection made, coz Connection is happens at Awaiting of instance
    # but subscription only registered after instance is already created, means already connected,
    # means On_connected event will never b called, it's can be twiked, but then Creating process of tester instance
    # will be less intuitive, and in one case subscription will work while in another not.

    def on_disconnected(self, callback: "Callable") -> None:
        """
        Register a callback which will be called at the time when connection will be closed.
        """

        self._conn.on_disconnected(callback)

    def on_reservation_change(self, callback: "Callable") -> None:
        """
        Register an callback function to C_RESERVATION event.
        """

        self._conn.subscribe(C_RESERVATION, callback)

    def on_reserved_by_change(self, callback: "Callable") -> None:
        """
        Register an callback function to C_RESERVEDBY event.
        """

        self._conn.subscribe(C_RESERVEDBY, callback)

    # endregion



import typing
from xoa_driver import ports
from xoa_driver.internals.hli_v1 import revisions
from xoa_driver.internals.utils.managers import ports_manager as pm

if typing.TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from .. import __interfaces as m_itf

from .module_l23_base import ModuleL23


@typing.final
@revisions.register_valkyrie_module(rev="Odin-10G-1S-2P")
class MOdin10G1S2P(ModuleL23):
    """Test module Odin-10G-1S-2P"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.POdin10G1S2P] = pm.PortsManager(
            conn=conn,
            ports_type=ports.POdin10G1S2P,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Odin-10G-1S-2P"""


@typing.final
@revisions.register_valkyrie_module(rev="Odin-10G-1S-2P[b]")
class MOdin10G1S2P_b(ModuleL23):
    """Test module Odin-10G-1S-2P[b]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.POdin10G1S2P_b] = pm.PortsManager(
            conn=conn,
            ports_type=ports.POdin10G1S2P_b,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Odin-10G-1S-2P[b]"""


@typing.final
@revisions.register_valkyrie_module(rev="Odin-10G-1S-2P[c]")
class MOdin10G1S2P_c(ModuleL23):
    """Test module Odin-10G-1S-2P[c]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.POdin10G1S2P_c] = pm.PortsManager(
            conn=conn,
            ports_type=ports.POdin10G1S2P_c,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Odin-10G-1S-2P[c]"""


@typing.final
@revisions.register_valkyrie_module(rev="Odin-10G-1S-6P")
class MOdin10G1S6P(ModuleL23):
    """Test module Odin-10G-1S-6P"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.POdin10G1S6P] = pm.PortsManager(
            conn=conn,
            ports_type=ports.POdin10G1S6P,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Odin-10G-1S-6P"""


@typing.final
@revisions.register_valkyrie_module(rev="Odin-10G-1S-6P[b]")
class MOdin10G1S6P_b(ModuleL23):
    """Test module Odin-10G-1S-6P[b]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.POdin10G1S6P_b] = pm.PortsManager(
            conn=conn,
            ports_type=ports.POdin10G1S6P_b,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Odin-10G-1S-6P[b]"""


@typing.final
@revisions.register_valkyrie_module(rev="Odin-10G-1S-2P-T")
class MOdin10G1S2PT(ModuleL23):
    """Test module Odin-10G-1S-2P-T"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.POdin10G1S2PT] = pm.PortsManager(
            conn=conn,
            ports_type=ports.POdin10G1S2PT,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Odin-10G-1S-2P-T"""


@typing.final
@revisions.register_valkyrie_module(rev="Odin-10G-1S-2P[d]")
class MOdin10G1S2P_d(ModuleL23):
    """Test module Odin-10G-1S-2P[d]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.POdin10G1S2P_d] = pm.PortsManager(
            conn=conn,
            ports_type=ports.POdin10G1S2P_d,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Odin-10G-1S-2P[d]"""


@typing.final
@revisions.register_valkyrie_module(rev="Odin-10G-1S-12P")
class MOdin10G1S12P(ModuleL23):
    """Test module Odin-10G-1S-12P"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.POdin10G1S12P] = pm.PortsManager(
            conn=conn,
            ports_type=ports.POdin10G1S12P,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager Odin-10G-1S-12P"""


@typing.final
@revisions.register_valkyrie_module(rev="Odin-10G-6S-6P[a]")
class MOdin10G6S6P_a(ModuleL23):
    """Test module Odin-10G-6S-6P[a]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.POdin10G6S6P_a] = pm.PortsManager(
            conn=conn,
            ports_type=ports.POdin10G6S6P_a,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager Odin-10G-6S-6P[a]"""


from __future__ import annotations
from typing import TYPE_CHECKING

from enum import (
    Enum,
    auto
)
from xoa_driver.internals.commands import enums
if TYPE_CHECKING:
    from xoa_driver.internals.commands import P_CAPABILITIES


class EPortInterfaceSubtype(Enum):
    Unspecified = auto()
    Single = auto()
    Dual = auto()
    Triple = auto()
    Quint = auto()


class SpeedDetector:
    """
    Work around of ports speed detection. Only for Genuine L23 ports.
    Keep it as a class for not couple all logic under one function.
    """
    __slots__ = ("capabilities", "interface")

    def __init__(self, capabilities: "P_CAPABILITIES.GetDataAttr", interface: str) -> None:
        self.capabilities = capabilities
        self.interface = interface

    @property
    def __determinate_interface_subtype(self) -> EPortInterfaceSubtype:
        if "[Triple]" in self.interface:
            return EPortInterfaceSubtype.Triple
        elif "[Quint]" in self.interface:
            return EPortInterfaceSubtype.Quint
        elif "[Dual]" in self.interface:
            return EPortInterfaceSubtype.Dual
        elif "[Single]" in self.interface:
            return EPortInterfaceSubtype.Single
        return EPortInterfaceSubtype.Unspecified

    @property
    def can_set_port_speed(self) -> bool:
        return self.__determinate_interface_subtype in {
            EPortInterfaceSubtype.Triple,
            EPortInterfaceSubtype.Quint,
            EPortInterfaceSubtype.Dual,
            EPortInterfaceSubtype.Single
        }

    def __define_single(self) -> list[enums.PortSpeedMode]:
        if "T1S" not in self.interface:
            return []
        return [
            enums.PortSpeedMode.AUTO,
            enums.PortSpeedMode.F100M,
            enums.PortSpeedMode.F1G,
        ]

    def __define_dual(self) -> list[enums.PortSpeedMode]:
        if "T1" in self.interface:
            return [enums.PortSpeedMode.F10MHDX, ]
        return []

    def __define_triple(self) -> list[enums.PortSpeedMode]:
        if self.interface.startswith("SFP"):
            if self.capabilities.max_speed == 1_000:
                return [
                    enums.PortSpeedMode.AUTO,
                    enums.PortSpeedMode.F10M,
                    enums.PortSpeedMode.F100M,
                    enums.PortSpeedMode.F1G,
                    enums.PortSpeedMode.F10M100M,
                    enums.PortSpeedMode.F100M1G,
                    enums.PortSpeedMode.F10MHDX,
                    enums.PortSpeedMode.F100MHDX,
                ]
            elif self.capabilities.max_speed == 10_000:
                return [
                    enums.PortSpeedMode.AUTO,
                    enums.PortSpeedMode.F100M,
                    enums.PortSpeedMode.F1G,
                    enums.PortSpeedMode.F10G,
                    enums.PortSpeedMode.F100M1G,
                ]
        elif self.interface.startswith("10GBASE-T"):
            return [
                enums.PortSpeedMode.AUTO,
                enums.PortSpeedMode.F100M,
                enums.PortSpeedMode.F1G,
                enums.PortSpeedMode.F10G,
                enums.PortSpeedMode.F100M1G,
            ]
        return []

    def __define_quint(self) -> list[enums.PortSpeedMode]:
        speeds = [
            enums.PortSpeedMode.AUTO,
            enums.PortSpeedMode.F100M,
            enums.PortSpeedMode.F1G,
            enums.PortSpeedMode.F2500M,
            enums.PortSpeedMode.F5G,
            enums.PortSpeedMode.F100M1G,
            enums.PortSpeedMode.F100M1G2500M,
        ]
        if self.interface.startswith("10GBASE-T"):
            speeds.append(enums.PortSpeedMode.F10G)
        return speeds

    def find_port_possible_speed(self) -> list[enums.PortSpeedMode]:
        if not self.can_set_port_speed:
            return []
        define_func = {
            EPortInterfaceSubtype.Single: self.__define_single,
            EPortInterfaceSubtype.Dual: self.__define_dual,
            EPortInterfaceSubtype.Triple: self.__define_triple,
            EPortInterfaceSubtype.Quint: self.__define_quint
        }.get(self.__determinate_interface_subtype, lambda: [])
        return define_func()



#: Other types which are used by ports or as a parameter to attribute set method.


from xoa_driver.internals.core.token import Token
from xoa_driver.internals.commands.subtypes import (
    ArpChunk,
    NdpChunk,
)
from xoa_driver.internals.core.transporter.protocol.payload.types import Hex

# indices types
from xoa_driver.internals.hli_v2.indices.connection_group.cg import ConnectionGroupIdx as ConnectionGroup
from xoa_driver.internals.hli_v2.indices.filter.base_filter import BaseFilterIdx as BasePortFilter
from xoa_driver.internals.hli_v2.indices.filter.genuine_filter import GenuineFilterIdx as GenuinePortFilter
from xoa_driver.internals.hli_v2.indices.length_term import LengthTermIdx as LengthTerm
from xoa_driver.internals.hli_v2.indices.match_term import MatchTermIdx as MatchTerm
from xoa_driver.internals.hli_v2.indices.port_dataset import PortDatasetIdx as PortDataset
from xoa_driver.internals.hli_v2.indices.streams.base_stream import BaseStreamIdx as BaseStream
from xoa_driver.internals.hli_v2.indices.streams.genuine_stream import GenuineStreamIdx as GenuineStream
from xoa_driver.internals.hli_v2.ports.port_l23.chimera.filter_definition.general import ProtocolSegment
from xoa_driver.internals.hli_v2.ports.port_l23.chimera.port_emulation import (
    CFlow as ImpairmentFlow,
    StatisticsTotals,
    CPerFlowStats as PerImpairmentFlowStats,
    CLatencyJitterImpairment,
    CDropImpairment,
    CMisorderingImpairment,
    CDuplicationImpairment,
    CCorruptionImpairment,
    CPolicerImpairment,
    CShaperImpairment,
)
from xoa_driver.internals.hli_v2.ports.port_l23.chimera.filter_definition.general import ModeBasic as BasicImpairmentFlowFilter
from xoa_driver.internals.hli_v2.ports.port_l23.chimera.filter_definition.general import ModeExtended as ExtendedImpairmentFlowFilter
from xoa_driver.internals.hli_v2.ports.port_l23.chimera.pe_custom_distribution import (
    CustomDistributions,
    CustomDistribution,
)
from xoa_driver.internals.hli_v2.ports.port_l23.chimera.filter_definition.shadow import (
    FilterDefinitionShadow,
    ModeExtendedS,
)
from xoa_driver.internals.hli_v2.ports.port_l23.chimera.filter_definition.general import ModeBasic


__all__ = (
    "Token",
    "ArpChunk",
    "NdpChunk",
    "Hex",
    "ConnectionGroup",
    "BasePortFilter",
    "GenuinePortFilter",
    "LengthTerm",
    "MatchTerm",
    "PortDataset",
    "BaseStream",
    "GenuineStream",
    "ImpairmentFlow",
    "BasicImpairmentFlowFilter",
    "ExtendedImpairmentFlowFilter",
    "StatisticsTotals",
    "CustomDistributions",
    "CustomDistribution",
    "PerImpairmentFlowStats",
    "CLatencyJitterImpairment",
    "CDropImpairment",
    "CMisorderingImpairment",
    "CDuplicationImpairment",
    "CCorruptionImpairment",
    "CPolicerImpairment",
    "CShaperImpairment",
    "FilterDefinitionShadow",
    "ModeExtendedS",
    "ModeBasic",
    "ProtocolSegment",
)



from enum import IntEnum
from typing import Final


NOTHING: Final = 0xFF
WILDCARD: Final = 0xFE
NOVALUE: Final = "N/A"
MAGIC_WORD: Final = b"XENA"
# MAGIC_WORD = 0x58454E41


def indices_format(number: int) -> str:
    return f"!{number}I"


class CommandType(IntEnum):
    """BXMP request type"""

    COMMAND_CHAIN = 1
    COMMAND_QUERY = 2
    COMMAND_VALUE = 3
    COMMAND_STATUS = 4


class CommandStatus(IntEnum):
    """Status codes in BXMP response message to request type <COMMAND_STATUS = 4>"""

    OK = 0
    """OK"""
    NOCONNECTIONS = 1
    """chassis has no available connection slots"""
    NOTLOGGEDON = 2
    """no command can be submitted before logon"""
    NOTRESERVED = 3
    """parameter cannot be set because containing resource not reserved"""
    NOTWRITABLE = 4
    """parameter cannot be set because it is not in a writable state"""
    NOTREADABLE = 5
    """parameter is write-only"""
    NOTVALID = 6
    """operation not valid in current state"""
    BADHEADER = 7
    """invalid magic word"""
    BADCOMMAND = 8
    """invalid command"""
    BADPARAMETER = 9
    """invalid parameter code"""
    BADMODULE = 10
    """invalid module index"""
    BADPORT = 11
    """invalid port index"""
    BADINDEX = 12
    """invalid parameter index"""
    BADSIZE = 13
    """invalid size of data"""
    BADVALUE = 14
    """invalid value of data"""
    FAILED = 15
    """failed to perform operation"""
    MEMORYFAILURE = 16
    """failed to allocate memory"""
    NOLICPE = 17
    """no free PE license"""
    NOLICPORT = 18
    """no free Port license"""
    NOTSUPPORTED = 19
    """feature not supported"""
    PENDING = 20
    """status return will wait until command has been executed"""
    MODULE_OPERATION_NOT_SUPPORTED_BY_CHASSIS = 21
    """module is not supported by chassis - e.g. because multi-image requires x64 OS."""
    XLSFAILED = 22
    """could not establish connection to Xena License Server"""
    XLSDENIED = 23
    """request for resource rejected by Xena License Server"""
    XLSINVALID = 24
    """Trying to run Valkyrie VE with VulcanVE-300 resource"""



from __future__ import annotations

import ctypes as c
from typing import (
    Type,
    ClassVar,
    TypeVar
)
from . import _constants as const


class ProtocolHeader(c.BigEndianStructure):
    """The structure for the header data of a packet in the protocol."""
    __slots__ = (
        "magic_word",
        "number_of_indices",
        "number_of_value_bytes",
        "command_parameter",
        "module_index",
        "port_index",
        "request_identifier",
    )

    _fields_ = [
        ("magic_word", c.c_char * 4),
        ("number_of_indices", c.c_ushort),
        ("number_of_value_bytes", c.c_ushort),
        ("command_parameter", c.c_ushort),
        ("module_index", c.c_ubyte),
        ("port_index", c.c_ubyte),
        ("request_identifier", c.c_uint32),
    ]

    @property
    def cmd_type(self) -> int:
        """Get command type.

        Returns:
            int: Command type
        """
        return (self.command_parameter & 0x0F00) >> 8

    @cmd_type.setter
    def cmd_type(self, val: int) -> None:
        """Set command type.

        Args:
            val (int): Command type
        """
        assert 0 <= val < 4
        self.command_parameter |= val << 8

    @property
    def cmd_code(self) -> int:
        """Get command code.

        Returns:
            int: Command code
        """
        last = (self.command_parameter & 0xF000) >> 4
        first = self.command_parameter & 0x00FF
        return first | last

    @cmd_code.setter
    def cmd_code(self, val: int) -> None:
        """Set command code.

        Args:
            val (int): Command code
        """
        first = (val & 0x0F00) << 4
        last = val & 0x00FF
        self.command_parameter |= first | last

    @property
    def body_size(self) -> int:
        """Calculate packet body size

        Returns:
            int: Packet body size
        """
        if self.magic_word != const.MAGIC_WORD:
            return 4
        body_size = self.number_of_value_bytes + (self.number_of_indices * 4)
        body_size += (4 - body_size % 4) % 4
        return body_size


Self = TypeVar("Self", bound="ResponseHeader")


class ResponseHeader(ProtocolHeader):
    """The structure for the response header data of a packet in the protocol."""
    size: ClassVar[int] = c.sizeof(ProtocolHeader)

    @property
    def is_pushed(self) -> bool:
        """Check if response is pushed."""
        return self.request_identifier == 0

    @classmethod
    def from_bytes(cls: Type[Self], buff: bytes) -> Self | None:
        """Create ResponseHeader object from bytes."""
        header_bytes = memoryview(buff)
        if not cls._header_bytes_is_valid(header_bytes, cls.size, const.MAGIC_WORD):
            return None
        return cls.from_buffer_copy(header_bytes)

    @staticmethod
    def _header_bytes_is_valid(header_bytes: memoryview, expected_size: int, magic_wrd: bytes) -> bool:
        """Check whether the header bytes are valid.

        Args:
            header_bytes (memoryview): Header bytes
            expected_size (int): Expected header size
            magic_wrd (bytes): Magic word bytes

        Returns:
            bool: True if the header bytes are valid, False otherwise
        """
        is_correct_size = len(header_bytes) == expected_size
        start_with_mw = header_bytes[:len(magic_wrd)] == magic_wrd
        return is_correct_size and start_with_mw



from __future__ import annotations
from dataclasses import dataclass
import typing

from xoa_driver.internals.core.builders import (
    build_get_request,
    build_set_request
)
from xoa_driver.internals.core import interfaces
from xoa_driver.internals.core.token import Token
from xoa_driver.internals.core.transporter.registry import register_command
from xoa_driver.internals.core.transporter.protocol.payload import (
    field,
    RequestBodyStruct,
    ResponseBodyStruct,
    XmpLong,
    XmpSequence,
)


@register_command
@dataclass
class PT_TOTAL:
    """
    Obtains statistics concerning all the packets transmitted on a port.
    """

    code: typing.ClassVar[int] = 230
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        bit_count_last_sec: int = field(XmpLong())
        """long integer, number of bits transmitted in the last second."""
        packet_count_last_sec: int = field(XmpLong())
        """long integer, number of packets transmitted in the last second."""
        byte_count_since_cleared: int = field(XmpLong())
        """long integer, number of bytes transmitted since statistics were cleared."""
        packet_count_since_cleared: int = field(XmpLong())
        """long integer, number of packets transmitted since statistics were cleared."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning all the packets transmitted on a port.

        :return: number of bits transmitted in the last second, number of packets transmitted in the last second,
            number of bytes transmitted since statistics were cleared, and number of packets transmitted since statistics were cleared.
        :rtype: PT_TOTAL.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PT_NOTPLD:
    """
    Obtains statistics concerning the packets without a test payload transmitted on
    a port.
    """

    code: typing.ClassVar[int] = 231
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        bit_count_last_sec: int = field(XmpLong())
        """long integer, number of bits transmitted in the last second."""
        packet_count_last_sec: int = field(XmpLong())
        """long integer, number of packets transmitted in the last second."""
        byte_count_since_cleared: int = field(XmpLong())
        """long integer, number of bytes transmitted since statistics were cleared."""
        packet_count_since_cleared: int = field(XmpLong())
        """long integer, number of packets transmitted since statistics were cleared."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning the packets without a test payload transmitted on a port.

        :return: number of bits transmitted in the last second, number of packets transmitted in the last second,
            number of bytes transmitted since statistics were cleared, and number of packets transmitted since statistics were cleared
        :rtype: PT_NOTPLD.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PT_STREAM:
    """
    Obtains statistics concerning the packets of a specific stream transmitted on a
    port.
    """

    code: typing.ClassVar[int] = 232
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _stream_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        bit_count_last_sec: int = field(XmpLong())
        """long integer, number of bits transmitted in the last second."""
        packet_count_last_sec: int = field(XmpLong())
        """long integer, number of packets transmitted in the last second."""
        byte_count_since_cleared: int = field(XmpLong())
        """long integer, number of bytes transmitted since statistics were cleared."""
        packet_count_since_cleared: int = field(XmpLong())
        """long integer, number of packets transmitted since statistics were cleared."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning the packets of a specific stream transmitted on a port.

        :return: number of bits transmitted in the last second, number of packets transmitted in the last second,
            number of bytes transmitted since statistics were cleared, and number of packets transmitted since statistics were cleared.
        :rtype: PT_STREAM.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._stream_xindex]))


@register_command
@dataclass
class PT_CLEAR:
    """
    Clear all the transmit statistics for a port. The byte and packet counts will
    restart at zero.
    """

    code: typing.ClassVar[int] = 233
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class SetDataAttr(RequestBodyStruct):
        pass

    def set(self) -> Token[None]:
        """Clear all the transmit statistics for a port. The byte and packet counts will restart at zero.
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PT_EXTRA:
    """
    Obtains additional statistics for packets transmitted on a port.
    """

    code: typing.ClassVar[int] = 235
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        tx_arp_req_count: typing.List[int] = field(XmpSequence(types_chunk=[XmpLong()]))
        """long integer, number of ARP requests transmitted"""

    def get(self) -> Token[GetDataAttr]:
        """Get additional statistics for packets transmitted on a port.

        :return: number of ARP requests transmitted
        :rtype: PT_EXTRA.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PT_TOTALEXT:
    """
    .. versionadded: v1.1

    An extension to :class:`PT_TOTAL` that also includes a calculation of bytes transmitted in the last second.
    It returns list of long integers; this list may be expanded in future software releases.
    """

    code: typing.ClassVar[int] = 236
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        bit_count_last_sec: int = field(XmpLong())
        """long integer, number of bits transmitted in the last second."""
        byte_count_last_sec: int = field(XmpLong())
        """long integer, number of bytes transmitted in the last second."""
        packet_count_last_sec: int = field(XmpLong())
        """long integer, number of packets transmitted in the last second."""
        byte_count_since_cleared: int = field(XmpLong())
        """long integer, number of bytes transmitted since statistics were cleared."""
        packet_count_since_cleared: int = field(XmpLong())
        """long integer, number of packets transmitted since statistics were cleared."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning all the packets transmitted on a port.

        :return: number of bits transmitted in the last second, number of bytes transmitted in the last second,
            number of packets transmitted in the last second, number of bytes transmitted since statistics were cleared,
            and number of packets transmitted since statistics were cleared.
        :rtype: PT_TOTALEXT.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PT_NOTPLDEXT:
    """
    .. versionadded: v1.1

    An extension to :class:`PT_NOTPLD` that also includes a calculation of bytes transmitted in the last second.
    It returns list of long integers; this list may be expanded in future software releases.
    """

    code: typing.ClassVar[int] = 237
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        bit_count_last_sec: int = field(XmpLong())
        """long integer, number of bits transmitted in the last second."""
        byte_count_last_sec: int = field(XmpLong())
        """long integer, number of bytes transmitted in the last second."""
        packet_count_last_sec: int = field(XmpLong())
        """long integer, number of packets transmitted in the last second."""
        byte_count_since_cleared: int = field(XmpLong())
        """long integer, number of bytes transmitted since statistics were cleared."""
        packet_count_since_cleared: int = field(XmpLong())
        """long integer, number of packets transmitted since statistics were cleared."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning the packets without a test payload transmitted on a port.

        :return: number of bits transmitted in the last second, number of bytes transmitted in the last second,
            number of packets transmitted in the last second, number of bytes transmitted since statistics were cleared,
            and number of packets transmitted since statistics were cleared
        :rtype: PT_NOTPLDEXT.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class PT_STREAMEXT:
    """
    .. versionadded: v1.1

    An extension to :class:`PT_STREAM` that also includes a calculation of bytes transmitted in the last second.
    It returns list of long integers; this list may be expanded in future software releases.
    """

    code: typing.ClassVar[int] = 238
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _stream_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        bit_count_last_sec: int = field(XmpLong())
        """long integer, number of bits transmitted in the last second."""
        byte_count_last_sec: int = field(XmpLong())
        """long integer, number of bytes transmitted in the last second."""
        packet_count_last_sec: int = field(XmpLong())
        """long integer, number of packets transmitted in the last second."""
        byte_count_since_cleared: int = field(XmpLong())
        """long integer, number of bytes transmitted since statistics were cleared."""
        packet_count_since_cleared: int = field(XmpLong())
        """long integer, number of packets transmitted since statistics were cleared."""

    def get(self) -> Token[GetDataAttr]:
        """Get statistics concerning the packets of a specific stream transmitted on a port.

        :return: number of bits transmitted in the last second, number of bytes transmitted in the last second,
            number of packets transmitted in the last second, number of bytes transmitted since statistics were cleared,
            and number of packets transmitted since statistics were cleared.
        :rtype: PT_STREAMEXT.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._stream_xindex]))


@register_command
@dataclass
class PT_FLOWTOTAL:
    """
    Obtains statistics concerning all the packets transmitted from a between this
    receive port and its partner TX port.
    """

    code: typing.ClassVar[int] = 1740
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        l2_bps: int = field(XmpLong())
        """long integer, number of bits transmitted at layer 2 in the last second for the flow."""
        pps: int = field(XmpLong())
        """long integer, number of packets transmitted in the last second for the flow."""
        byte_count: int = field(XmpLong())
        """long integer, number of bytes transmitted since statistics were cleared for the flow."""
        packet_count: int = field(XmpLong())
        """long integer, number of packets transmitted since statistics were cleared for the flow."""

    def get(self) -> Token[GetDataAttr]:
        """(Chimera only) Get statistics concerning all the packets transmitted from a between this receive port and its partner TX port.

        :return:
            number of bits transmitted at layer 2 in the last second for the flow,
            number of packets transmitted in the last second for the flow,
            number of bytes transmitted since statistics were cleared for the flow,
            number of packets transmitted since statistics were cleared for the flow

        :rtype: PT_FLOWTOTAL.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))


@register_command
@dataclass
class PT_FLOWCLEAR:
    """
    Clear all the transmit statistics on a particular flow for a Chimera port. The
    byte and packet counts will restart at zero.
    """

    code: typing.ClassVar[int] = 1742
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class SetDataAttr(RequestBodyStruct):
        pass

    def set(self) -> Token[None]:
        """(Chimera only) Clear all the transmit statistics on a particular flow for a Chimera port. The byte and packet counts will restart at zero.
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))



from __future__ import annotations
import struct
from typing import Any
from . import _constants as const
from . import _utils
from .struct_header import ResponseHeader
from .payload import ResponseBodyStruct
from .exceptions import get_status_error, XmpStatusException


class Response:
    """
    Response serializer
    """

    __slots__ = (
        "class_name",
        "header",
        "index_values",
        "values",
        "__buffer",
    )

    def __init__(self, class_name: str, header: ResponseHeader, buffer: bytes, response_struct: type[ResponseBodyStruct] | None) -> None:
        self.class_name = class_name
        self.header = header
        idces_fmt_ = const.indices_format(header.number_of_indices)
        idx_count_ = struct.calcsize(idces_fmt_)
        self.__buffer = memoryview(buffer)
        self.index_values = self.__parse_indices(idces_fmt_, self.__buffer[:idx_count_])
        payload_position = slice(idx_count_, idx_count_ + header.number_of_value_bytes)
        self.values: Any = self.__parse_values(self.__buffer[payload_position], response_struct)

    def __str__(self) -> str:
        return _utils.format_str(self)

    def __repr__(self) -> str:
        return _utils.format_repr(self)

    def __bytes__(self) -> bytes:
        return bytes().join((self.header, self.__buffer))

    def __parse_indices(self, fmt: str, buffer: memoryview) -> list[int]:
        return list(struct.unpack_from(fmt, buffer, 0))

    def __parse_values(self, buffer: memoryview, struct_type: type[ResponseBodyStruct] | None) -> ResponseBodyStruct | None:
        if self.header.cmd_type == const.CommandType.COMMAND_VALUE:
            return struct_type(buffer) if struct_type else None
        return None

    def get_error(self) -> XmpStatusException | None:
        if self.is_ok:
            return None
        exception = get_status_error(self.command_status)
        return exception(self.class_name)

    @property
    def is_pushed(self) -> bool:
        return self.header.is_pushed

    @property
    def request_identifier(self) -> int:
        return self.header.request_identifier

    @property
    def cmd_code(self) -> int:
        return self.header.cmd_code

    @property
    def command_status(self) -> const.CommandStatus | None:
        if self.header.cmd_type == const.CommandType.COMMAND_STATUS:
            return const.CommandStatus(self.header.cmd_code)
        return None

    @property
    def is_ok(self) -> bool:
        return self.get_return_ok or self.set_return_ok

    @property
    def set_return_ok(self) -> bool:
        is_status_resp = self.header.cmd_type == const.CommandType.COMMAND_STATUS
        is_succesful = self.header.cmd_code == const.CommandStatus.OK
        return is_status_resp and is_succesful

    @property
    def get_return_ok(self) -> bool:
        is_payload_resp = self.header.cmd_type == const.CommandType.COMMAND_VALUE
        contain_values = self.values is not None
        return is_payload_resp and contain_values



from __future__ import annotations
from typing import Type
from ._typings import XoaCommandType


class XmpRegistryException(Exception):
    ...


class XmpCmdDuplicatedDefenitionError(XmpRegistryException):
    def __init__(self, xmc_cls: Type[XoaCommandType]) -> None:
        self.cmd_type = xmc_cls
        self.msg = f"The defenition of <{self.cmd_type.__name__}> is duplicated."
        super().__init__(self.msg)


class XmpCmdNotImplemented(XmpRegistryException):
    def __init__(self, cmd_idx: int) -> None:
        self.cmd_idx = cmd_idx
        self.msg = f"The command of the code: [{cmd_idx}], is not implemented or not registered as a XMP command."
        super().__init__(self.msg)


COMMANDS_REGISTRY: dict[int, Type[XoaCommandType]] = dict()
"""A global registry of which map command id to its class implementation."""


def register_command(xmc_cls: Type[XoaCommandType]) -> Type:
    """A decorator used for register commands descriptors in to global registry"""
    global COMMANDS_REGISTRY
    if xmc_cls.code in COMMANDS_REGISTRY:
        raise XmpCmdDuplicatedDefenitionError(xmc_cls)
    COMMANDS_REGISTRY[xmc_cls.code] = xmc_cls
    return xmc_cls


def get_command(command_idx: int) -> Type[XoaCommandType]:
    """Method allows to get a command class by the command code"""
    global COMMANDS_REGISTRY
    xmc_type = COMMANDS_REGISTRY.get(command_idx, None)
    if not xmc_type:
        raise XmpCmdNotImplemented(command_idx)
    return xmc_type



from __future__ import annotations
import sys
import typing as t

if t.TYPE_CHECKING:
    try:
        from loguru import Logger
    except ImportError:
        pass
    from .__logger import TransportationLogger


SYMBOL_REQUEST = "<G><w> -> </w></G>"
SYMBOL_PUSH = "<M><w> -P </w></M>"
SYMBOL_RESPONSE = "<Y><e> <- </e></Y>"


class StateOnLoguru:

    @staticmethod
    def setup(connid: str, logger: "Logger") -> "Logger":
        logger.remove()  # All configured handlers are removed
        fmt = "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <level>{message}</level>"
        logger.add(sys.stdout, format=fmt)
        logger = logger.bind(connid=connid)
        return logger

    @staticmethod
    def info(inst: "TransportationLogger", msg: t.Any) -> None:
        _loger = t.cast("Logger", inst._logger)
        _loger.opt(lazy=True, colors=True).info(msg)

    @staticmethod
    def error(inst: "TransportationLogger", msg: t.Any) -> None:
        _loger = t.cast("Logger", inst._logger)
        _loger.opt(lazy=True, colors=True).error(msg)

    @staticmethod
    def debug_request(inst: "TransportationLogger", request: object) -> None:
        _loger = t.cast("Logger", inst._logger)
        _loger.opt(lazy=True, colors=True).debug(f"{SYMBOL_REQUEST} {request!r}")

    @staticmethod
    def debug_response(inst: "TransportationLogger", response: object) -> None:
        _loger = t.cast("Logger", inst._logger)
        _loger.opt(lazy=True, colors=True).debug(f"{SYMBOL_RESPONSE} {response!r}")

    @staticmethod
    def debug_push(inst: "TransportationLogger", response: object) -> None:
        _loger = t.cast("Logger", inst._logger)
        _loger.opt(lazy=True, colors=True).debug(f"{SYMBOL_PUSH} {response!r}")
